#!/usr/bin/env python
# coding=utf-8

import cmd
import inspect
import io
import os
import multiprocessing
import re
import time
import signal
import shutil
import sys
import warnings
import yaml

from tempfile import gettempdir

from assemblyline.common import forge, log as al_log
from assemblyline.common.backupmanager import DistributedBackup
from assemblyline.common.security import get_totp_token, generate_random_secret
from assemblyline.common.uid import get_random_id
from assemblyline.common.version import FRAMEWORK_VERSION, SYSTEM_VERSION
from assemblyline.odm.models.signature import RULE_STATUSES
from assemblyline.remote.datatypes.hash import Hash

warnings.filterwarnings("ignore")

config = forge.get_config()
config.logging.log_to_console = False
al_log.init_logging('cli')

PROCESSES_COUNT = 50
COUNT_INCREMENT = 500
DATASTORE = None
t_count = 0
t_last = time.time()


class NullLogger(object):
    @staticmethod
    def info(msg, *args, **kwargs):
        pass

    @staticmethod
    def warning(msg, *args, **kwargs):
        pass

    @staticmethod
    def warn(msg, *args, **kwargs):
        pass

    @staticmethod
    def error(msg, *args, **kwargs):
        pass

    @staticmethod
    def exception(msg, *args, **kwargs):
        pass


class PrintLogger(object):
    @staticmethod
    def info(msg, end=None):
        print(msg, end=end)

    @staticmethod
    def warning(msg, end=None):
        print(f"[W] {msg}", end=end)

    @staticmethod
    def warn(msg, end=None):
        print(f"[W] {msg}", end=end)

    @staticmethod
    def error(msg, end=None):
        print(f"[E] {msg}", end=end)

    @staticmethod
    def exception(msg, end=None):
        print(f"[EX] {msg}", end=end)


def init():
    global DATASTORE
    DATASTORE = forge.get_datastore(archive_access=True)
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def submission_delete_tree(key, logger):
    try:
        with forge.get_filestore() as f_transport:
            # noinspection PyUnresolvedReferences
            DATASTORE.delete_submission_tree(key, transport=f_transport)
    except Exception as e:
        logger.error(e)
        return "DELETE", "submission", key, False, isinstance(logger, PrintLogger)

    return "deleted", "submission", key, True, isinstance(logger, PrintLogger)


def action_done(args):
    global t_count, t_last, COUNT_INCREMENT
    action, bucket, key, success, do_print = args
    if success:
        t_count += 1
        if t_count % COUNT_INCREMENT == 0:
            new_t = time.time()
            if do_print:
                print(f"[{bucket}] {t_count} {action} so far ({new_t - t_last}"
                      f" at {int(COUNT_INCREMENT / (new_t - t_last))} keys/sec)")
            t_last = new_t
    elif do_print:
        print(f"!!ERROR!! [{bucket}] {action} ==> {key}")


# noinspection PyMethodMayBeStatic,PyProtectedMember,PyBroadException
class ALCommandLineInterface(cmd.Cmd):  # pylint:disable=R0904

    def __init__(self, show_prompt=True, logger_class=PrintLogger):
        cmd.Cmd.__init__(self)
        self.logger = logger_class()
        self.prompt = ""
        self.intro = ""
        self.datastore = forge.get_datastore(archive_access=True)
        self.config = forge.get_config()
        if show_prompt:
            self._update_context()
        self._platform = sys.platform
        self._cmd_lex = None

    def _update_context(self):
        self.prompt = '(al_cli) $ '
        self.intro = 'AL 4.0 - Console. (type help).'

    def _compile_cmd_lex(self):
        """Compiles regular expression for for command-line splitting.
        For use with subprocess, for argv injection etc. Using fast REGEX.

        self._platform: auto from current platform;
                        'win32' = Windows/CMD
                        'linux', etc. = POSIX
        """

        if self._platform != 'win32':
            self._cmd_lex = re.compile(r'''"
            ((?:\\["\\]|[^"])*)"|       # quoted string
            '([^']*)'|                  # quoted single string
            (\\.)|                      # escaped string
            (&&?|\|\|?|\d?>|[<])|      # pipes and other command continuation
            ([^\s'"\\&|<>]+)|           # words
            (\s+)|                      # whitespace
            (.)                         # fail
            ''', re.VERBOSE)
        else:
            self._cmd_lex = re.compile(r'''
            "((?:""|\\["\\]|[^"])*)"|   # quoted string
            (\\\\(?=\\*")|\\")|         # escaped string
            (&&?|\|\|?|\d?>|[<])|       # pipes and other command continuation
            ([^\s"&|<>]+)|              # words
            (\s+)|                      # whitespace
            (.)                         # fail
            ''', re.VERBOSE)

    def _parse_args(self, s):
        """Multi-platform variant of shlex.split() for command-line splitting.
        For use with subprocess, for argv injection etc. Using fast compiled class REGEX.
        """
        if self._cmd_lex is None:
            self._compile_cmd_lex()

        args = []
        accu = None  # collects pieces of one arg
        for qs, qss, esc, pipe, word, white, fail in self._cmd_lex.findall(s):
            if word:
                pass  # most frequent
            elif esc:
                word = esc[1]
            elif white or pipe:
                if accu is not None:
                    args.append(accu)
                if pipe:
                    args.append(pipe)
                accu = None
                continue
            elif fail:
                raise ValueError("invalid or incomplete shell string")
            elif qs:
                word = qs.replace('\\"', '"').replace('\\\\', '\\')
                if self._platform == "win32":
                    word = word.replace('""', '"')
            else:
                word = qss  # may be even empty; must be last

            accu = (accu or '') + word

        if accu is not None:
            args.append(accu)

        return args

    def _get_completion_list(self, text, line, collection_name,
                             valid_actions=None, terminal_actions=None, multiple_actions=None):
        """
        Helper for command completion for subcommands that take no arguments, take one argument,
        or multiple arguments of the same type.
        """
        if terminal_actions is None:
            terminal_actions = []
        if valid_actions is None:
            valid_actions = []
        if multiple_actions is None:
            multiple_actions = []

        args = self._parse_args(line)
        if text == '':
            # Because _parse_args strips trailing spaces
            args.append(text)

        if len(args) == 2:
            return [i for i in valid_actions if i.startswith(text)]
        elif len(args) >= 3:
            if args[1] in terminal_actions:
                return []
            collection = self.datastore.get_collection(collection_name)
            if collection:
                if len(args) == 3:
                    return [i for i in collection.keys() if i.startswith(text)]
                if args[1] in multiple_actions:
                    return [i for i in collection.keys() if i.startswith(text) and i not in args[2:-1]]
        return []

    def _print_error(self, msg):
        stack_func = None
        stack = inspect.stack()
        for item in stack:
            if 'cli.py' in item[1] and '_print_error' not in item[3]:
                stack_func = item[3]
                break

        if msg:
            self.logger.error(f"{msg}\n")

        if stack_func:
            function_doc = inspect.getdoc(getattr(self, stack_func))
            if function_doc:
                self.logger.info(f"Function help:\n\n{function_doc}\n")

    #
    # Exit actions
    #
    def do_exit(self, arg):
        """Quits the CLI"""
        arg = arg or 0
        sys.exit(int(arg))

    def do_quit(self, arg):
        """Quits the CLI"""
        self.do_exit(arg)

    #
    # Backup actions
    #
    def do_backup(self, args):
        """
        Backup the database content to a set of json files

        Usage:
            backup <destination_folder>
                   <destination_folder> <bucket_name> [follow] [force] <query>

        Parameters:
            <destination_folder> Path to the destination folder [required]

            <bucket_name>        Name of the bucket to backup [required in backup by query]

            follow               Follow IDs to backup more then the specified bucket
                                 [optional, only used in backup by query]

            force                Automatically perform backup without asking for confirmation
                                 [optional, only used in backup by query]

            <query>              Query that the data need to match
                                 [optional, only used in backup by query]

        Examples:
            # Create a backup of the system buckets
            backup /tmp/backup_folder

            # Created a backup of all alerts
            backup /tmp/alerts_backup alert "*:*"
        """
        args = self._parse_args(args)

        follow = False
        if 'follow' in args:
            follow = True
            args.remove('follow')

        force = False
        if 'force' in args:
            force = True
            args.remove('force')

        if len(args) == 1:
            dest = args[0]
            system_backup = True
            bucket = None
            follow = False
            query = None
        elif len(args) == 3:
            dest, bucket, query = args
            system_backup = False
        else:
            self._print_error("Wrong number of arguments for backup command.")
            return

        if os.path.exists(dest):
            self.logger.error("Use a different backup directory, this one already exists.")
            return

        try:
            os.makedirs(dest, exist_ok=True)
        except Exception:
            temp_dir = gettempdir()
            self.logger.error(f"Cannot make {dest!s} folder. Make sure you can write to this folder.\n"
                              f"Maybe you should write your backups in {temp_dir} ?")
            return

        if system_backup:
            system_buckets = [
                'heuristic',
                'service',
                'service_delta',
                'signature',
                'user',
                'user_avatar',
                'user_favorites',
                'user_settings',
                'workflow'
            ]

            backup_manager = DistributedBackup(dest, worker_count=5, logger=self.logger)

            try:
                backup_manager.backup(system_buckets)
            except KeyboardInterrupt:
                backup_manager.cleanup()
                raise
        else:
            data = self.datastore.get_collection(bucket).search(query, offset=0, rows=1)
            total = data['total']
            if not total:
                self.logger.info(f"\nNothing in '{bucket.upper()}' matches the query:\n\n  {query}\n")
                try:
                    os.rmdir(dest)
                except Exception:
                    self.logger.error(f"Cannot remove backup destination folder '{dest}'.")
                return
            else:
                self.logger.info(f"\nNumber of items matching this query: {data['total']}\n")

            if not force:
                self.logger.info("This is an example of the data that will be backed up:\n")
                self.logger.info(f"{data['items'][0]}\n")
                if self.prompt:
                    cont = input("Are your sure you want to continue? (y/N) ")
                    cont = cont == "y"
                else:
                    self.logger.warn("You are not in interactive mode therefor the backup was not executed. "
                                     "Add 'force' to your commandline to execute the backup.")
                    cont = False

                if not cont:
                    self.logger.warn("\n**ABORTED**\n")
                    try:
                        os.rmdir(dest)
                    except Exception:
                        self.logger.error(f"Cannot remove backup destination folder '{dest}'.")
                    return

            if follow:
                total *= 100

            worker_count = int(max(2, min(total / 1000, multiprocessing.cpu_count() * 2)))
            backup_manager = DistributedBackup(dest, worker_count=worker_count, logger=self.logger)

            try:
                backup_manager.backup([bucket], follow_keys=follow, query=query)
            except KeyboardInterrupt:
                backup_manager.cleanup()
                raise

    def do_restore(self, args):
        """
        Restore a backup created by the backup command

        Usage:
            restore <backup_directory>

        Parameters:
            <backup_directory> Path to the backup folder [required]

        Examples:
            restore /tmp/backup_folder
        """
        args = self._parse_args(args)

        if len(args) not in [1]:
            self._print_error("Wrong number of arguments for restore command.")
            return

        path = args[0]
        if not path:
            self._print_error("You must specify an input folder.")
            return

        workers = len([x for x in os.listdir(path) if '.part' in x])

        # Make sure that all the backup parts are there
        bak_files = os.listdir(path)
        for part_num in range(workers):
            suffix_check = f".part{part_num:d}"
            if not any(suffix_check in x for x in bak_files):
                self._print_error(f"{workers:d} files exist, but no file ending with {suffix_check} found. ")
                return

        backup_manager = DistributedBackup(path, worker_count=workers, logger=self.logger)

        try:
            backup_manager.restore()
        except KeyboardInterrupt:
            backup_manager.cleanup()
            raise

    def do_delete(self, args):
        """
        Delete all data from a bucket that match a given query

        Usage:
            delete <bucket> [full] [force] <query>

        Parameters:
            <bucket>  Name of the bucket to delete from
            full      Follow IDs and remap classification
                      [Optional: Only work while deleting submissions]
            force     Automatically perform deletion without asking for confirmation [optional]
            <query>   Query to run to find the data to delete

        Examples:
            # Delete all submission for "user" with all associated results
            delete submission full "params.submitter:user"
        """
        valid_buckets = list(self.datastore.ds.get_models().keys())
        args = self._parse_args(args)

        if 'full' in args:
            full = True
            args.remove('full')
        else:
            full = False

        if 'force' in args:
            force = True
            args.remove('force')
        else:
            force = False

        if len(args) != 2:
            self._print_error("Wrong number of arguments for delete command.")
            return

        bucket, query = args

        if bucket not in valid_buckets:
            bucket_list = '\n'.join(valid_buckets)
            self._print_error(f"\nInvalid bucket specified: {bucket}\n\nValid buckets are:\n{bucket_list}")
            return
        else:
            collection = self.datastore.get_collection(bucket)

        pool = None
        try:
            cont = force
            test_data = collection.search(query, offset=0, rows=1)
            if not test_data["total"]:
                self.logger.info("Nothing matches the query.")
                return

            if not force:
                self.logger.info(f"\nNumber of items matching this query: {test_data['total']}\n\n")
                self.logger.info("This is an example of the data that will be deleted:\n")
                self.logger.info(f"{test_data['items'][0]}\n")
                if self.prompt:
                    cont = input("Are your sure you want to continue? (y/N) ")
                    cont = cont == "y"

                    if not cont:
                        self.logger.warn("\n**ABORTED**\n")
                        return
                else:
                    self.logger.warn("You are not in interactive mode, therefore the delete was not executed. "
                                     "Add 'force' to your commandline to execute the delete.")
                    return

            if cont:
                if full and bucket == 'submission':
                    pool = multiprocessing.Pool(processes=PROCESSES_COUNT, initializer=init)
                    for data in collection.stream_search(query, fl="id", item_buffer_size=COUNT_INCREMENT):
                        pool.apply_async(submission_delete_tree, (data.id, self.logger), callback=action_done)
                else:
                    collection.delete_by_query(query)

        except KeyboardInterrupt:
            self.logger.warn("Interrupting jobs...")
            if pool is not None:
                pool.terminate()
                pool.join()

        except Exception as e:
            self.logger.exception(f"Something went wrong, retry!\n\n {e}\n")
        else:
            if pool is not None:
                pool.close()
                pool.join()
            collection.commit()
            self.logger.info(f"Data of bucket '{bucket}' matching query '{query}' has been deleted.")

    def do_service(self, args):
        """
        Perform operation on the different services on the system

        Usage:
            service list
                    cleanup
                    show    <name>
                    disable <name>
                    enable  <name>
                    remove  <name>

        Actions:
            list      List all services on the system
            cleanup   Remove services incompatible with your version
                      and fix all service deltas
            disable   Disable the service in the system
            enable    Enable the service in the system
            remove    Remove the service from the system
            show      Show the service description

        Parameters:
            <name>   Name of the service to perform the action on

        Examples:
            # Show service 'Extract'
            service show Extract
        """
        valid_actions = ['list', 'cleanup', 'show', 'disable', 'enable', 'remove']
        args = self._parse_args(args)

        if len(args) == 1:
            action_type = args[0]
            item_id = None
        elif len(args) == 2:
            action_type, item_id = args
        else:
            self._print_error("Wrong number of arguments for service command.")
            return

        if action_type not in valid_actions:
            self._print_error("Invalid action for service command.")
            return

        collection = self.datastore.get_collection('service_delta')

        if action_type == 'list':
            for key in collection.keys():
                self.logger.info(key)
            return
        elif action_type == 'cleanup':
            self.logger.info("Validating services deltas...")
            versions = set()
            for key in collection.keys():
                self.logger.info(f"\t{key}")
                svc_data = collection.get(key)
                versions.add(f"{key}_{svc_data.version}")
                collection.save(key, svc_data)

            self.logger.info("Validating installed services...")
            system_version = f"{FRAMEWORK_VERSION}.{SYSTEM_VERSION}."
            to_del = []
            svc_col = self.datastore.get_collection('service')
            for key in svc_col.keys():
                svc_data = svc_col.get(key)
                if not svc_data.version.startswith(system_version) and key not in versions:
                    to_del.append(key)
                else:
                    svc_col.get(key, svc_data)

            self.logger.info("Removing services not matching your system...")
            for svc_vers in to_del:
                self.logger.info(f"\t{svc_vers}")
                svc_col.delete(svc_vers)
            self.logger.info("Done!")
            return

        if item_id:
            item = collection.get(item_id)
            if item is None:
                self.logger.warn(f"{item_id} does not exist")
                return
        else:
            self._print_error("Invalid command parameters")
            return

        if action_type == 'show':
            output = io.StringIO()
            yaml.safe_dump(self.datastore.get_service_with_delta(item_id, as_obj=False), output)
            self.logger.info(output.getvalue())
        elif action_type == 'disable':
            item.enabled = False
            collection.save(item_id, item)
            self.logger.info(f"{item_id} was disabled")
        elif action_type == 'enable':
            item.enabled = True
            collection.save(item_id, item)
            self.logger.info(f"{item_id} was enabled")
        elif action_type == 'remove':
            collection.delete(item_id)
            service = self.datastore.get_collection('service')
            service.delete_by_query(f"name:{item_id}")
            self.logger.info(f"Service '{item_id}' removed.")

    def complete_service(self, text, line, begidx, endidx):
        """
        Command completion for the 'service' command
        """
        valid_actions = ['list', 'show', 'disable', 'enable', 'remove']
        terminal_actions = ['list']
        multiple_actions = []

        return self._get_completion_list(text, line, 'service_delta', valid_actions, terminal_actions, multiple_actions)

    def do_signature(self, args):
        """
        Perform operation on a signature in the system

        Usage:
            signature change_status by_id    [force] <status_value> <id>
                      change_status by_query [force] <status_value> <query>
                      remove        <id>
                      show          <id>

        Actions:
            change_status  Change the status of a signature
            remove         Remove the signature from the system
            show           Show the signature description

        Parameters:
            <id>           ID of the signature to perform the action on
            <query>        Query to match the signature
            <status_value> New status value for the signature
            force          Automatically perform status change without asking for confirmation [optional]
            by_id          Use an ID to choose the signature
            by_query       Use a search query to choose the signatures

        Examples:
            # Change the status of all STAGING signatures to DEPLOYED
            signature change_status by_query DEPLOYED "meta.al_status:STAGING"
        """
        valid_actions = ['show', 'change_status', 'remove']
        args = self._parse_args(args)

        if 'force' in args:
            force = True
            args.remove('force')
        else:
            force = False

        if len(args) == 2:
            action_type, item_id = args
            id_type = status = None
        elif len(args) == 4:
            action_type, id_type, status, item_id = args
        else:
            self._print_error("Wrong number of arguments for signature command.")
            return

        if action_type not in valid_actions:
            self._print_error("Invalid action for signature command.")
            return

        signatures = self.datastore.get_collection('signature')

        if action_type == 'show' and item_id:
            output = io.StringIO()
            yaml.safe_dump(signatures.get(item_id, as_obj=False), output)
            self.logger.info(output.getvalue())
        elif action_type == 'change_status' and item_id and id_type and status:
            if status not in RULE_STATUSES:
                statuses = "\n".join(RULE_STATUSES)
                self._print_error(f"\nInvalid status for action 'change_status' of signature command."
                                  f"\n\nValid statuses are:\n{statuses}")
                return

            if id_type == 'by_id':
                signature = signatures.get(item_id)
                signature.status = status
                signatures.save(item_id, signature)
                self.logger.info(f"Signature '{item_id}' was changed to status {status}.")
            elif id_type == 'by_query':
                try:
                    cont = force
                    test_data = signatures.search(item_id, offset=0, rows=1)
                    if not test_data["total"]:
                        self.logger.info("Nothing matches the query.")
                        return

                    if not force:
                        self.logger.info(f'\nNumber of items matching this query: {test_data["total"]}\n\n')
                        self.logger.info("This is an example of the signatures that will change status:\n")
                        self.logger.info(f"{test_data['items'][0]}\n")
                        if self.prompt:
                            cont = input("Are your sure you want to continue? (y/N) ")
                            cont = cont == "y"

                            if not cont:
                                self.logger.warn("\n**ABORTED**\n")
                                return
                        else:
                            self.logger.warn("You are not in interactive mode, "
                                             "therefore the status change was not executed. "
                                             "Add 'force' to your commandline to execute the status change.")
                            return

                    if cont:
                        updated = signatures.update_by_query(item_id,
                                                             [(signatures.UPDATE_SET, 'status', status)])
                        self.logger.info(f"Signatures matching query '{item_id}' were changed "
                                         f"to status '{status}'. [{updated}]")

                except KeyboardInterrupt:
                    self.logger.warn("Interrupting jobs...")
                except Exception as e:
                    self.logger.error(f"Something went wrong, retry!\n\n {e}\n")
            else:
                self._print_error("Invalid action parameters for action 'change_status' of signature command.")

        elif action_type == 'remove' and item_id:
            signatures.delete(item_id)
            self.logger.info(f"Signature '{item_id}' removed.")
        else:
            self._print_error("Invalid command parameters")

    def do_user(self, args):
        """
        Perform operation on a user in the system

        Usage:
            user list
                 show        <uname>
                 disable     <uname>
                 enable      <uname>
                 set_admin   <uname>
                 unset_admin <uname>
                 remove      <uname>
                 set_otp     <uname>
                 unset_otp   <uname>
                 show_otp    <uname>

        Actions:
            list         List all the users
            show         Describe a user
            disable      Disable a user
            enable       Enable a user
            set_admin    Grant a user admin privileges
            unset_admin  Revoke a user's admin privileges
            remove       Remove a user
            set_otp      Generate a new random OTP secret key for user
            unset_otp    Remove OTP Secret Token
            show_otp     Show current OTP Token

        Parameters:
            <uname>      Username of the user to perform the action on


        Examples:
            # Disable user 'user'
            user disable user
        """
        valid_actions = ['list', 'show', 'disable', 'enable', 'remove',
                         'set_admin', 'unset_admin', 'set_otp', 'unset_otp', 'show_otp']
        args = self._parse_args(args)

        if len(args) == 1:
            action_type = args[0]
            item_id = None
        elif len(args) == 2:
            action_type, item_id = args
        else:
            self._print_error("Wrong number of arguments for user command.")
            return

        if action_type not in valid_actions:
            self._print_error("Invalid action for user command.")
            return

        users = self.datastore.get_collection('user')

        if action_type == 'list':
            for key in users.keys():
                self.logger.info(key)
            return
        if item_id:
            item = users.get(item_id)
            if item is None:
                self.logger.warn(f"{item_id} does not exist")
                return
        else:
            self._print_error("Invalid command parameters")
            return

        if action_type == 'show':
            output = io.StringIO()
            yaml.safe_dump(item.as_primitives(), output)
            self.logger.info(output.getvalue())
        elif action_type == 'disable':
            item.is_active = False
            users.save(item_id, item)
            self.logger.info(f"{item_id} was disabled")
        elif action_type == 'enable':
            item.is_active = True
            users.save(item_id, item)
            self.logger.info(f"{item_id} was enabled")
        elif action_type == 'set_admin':
            if 'admin' not in item.type:
                item.type.append('admin')
                users.save(item_id, item)
            self.logger.info(f"Granted admin privileges to {item_id}")
        elif action_type == 'unset_admin':
            if 'admin' in item.type:
                item.type.remove('admin')
                users.save(item_id, item)
            self.logger.info(f"Admin privileges revoked for {item_id}")
        elif action_type == 'remove':
            users.delete(item_id)
            self.logger.info(f"User '{item_id}' removed.")
        elif action_type == 'unset_otp':
            item.otp_sk = None
            users.save(item_id, item)
            self.logger.info(f"{item_id} OTP secret key was removed")
        elif action_type == 'set_otp':
            item.otp_sk = generate_random_secret()
            users.save(item_id, item)
            self.logger.info(f"{item_id} OTP secret key is now: {item.otp_sk}")
        elif action_type == 'show_otp':
            if item.otp_sk:
                try:
                    while True:
                        self.logger.info('\r{!s} OTP Token:   {:06d}   {:░<30}'.format(
                            item_id,
                            get_totp_token(item.otp_sk),
                            "█" * int(time.time() % 30)),
                            end=''
                        ),
                        sys.__stdout__.flush()
                        time.sleep(1)
                except KeyboardInterrupt:
                    self.logger.info('')
            else:
                self.logger.warn(f"2FA not enabled for user {item_id}")

    def complete_user(self, text, line, begidx, endidx):
        """
        Command completion for the 'user' command
        """
        valid_actions = ['list', 'show', 'disable', 'enable', 'remove',
                         'set_admin', 'unset_admin', 'set_otp', 'unset_otp', 'show_otp']
        terminal_actions = ['list']
        multiple_actions = []

        return self._get_completion_list(text, line, 'user', valid_actions, terminal_actions, multiple_actions)

    def do_index(self, args):
        """
        Perform operations on the database index.

        ** Do not use these operations unless you absolutely have to as they may slow down
           considerably your system and some of these operations may result in dataloss if
           something went wrong in the middle of it.

        Usage:
            index commit        [<safe>] [<bucket>]
                  reindex       [<safe>] [<bucket>]
                  fix_ilm       [<safe>] [<bucket>]
                  fix_replicas  [<safe>] [<bucket>]
                  fix_shards    [<safe>] [<bucket>]

        Actions:
            commit        Force datastore to commit the specified index
            reindex       Force a reindex of the sepcified index
                             ** This operation is really slow because it re-index all documents
            fix_ilm       Fix ILM on specified indices
                             ** This operation can be really slow when going from an ILM setup to a hot
                                archive only setup because it will copy the archive to hot index
            fix_replicas  Fix replica count on specified indices
            fix_shards    Fix sharding on specified indices
                             ** This operation can be slow and will prevent data from being written
                                the cluster while it is hapenning.

        Parameters:
            <safe>       Does not validate the model [optional]
            <bucket>     Bucket to do the operation on [optional]


        Examples:
            # Force commit on file bucket
            index commit file
            # Force commit on all bucket
            index commit
        """

        valid_buckets = list(self.datastore.ds.get_models().keys())
        valid_actions = ['commit', 'reindex', 'fix_shards', 'fix_ilm', 'fix_replicas']

        args = self._parse_args(args)

        safe = False
        if 'safe' in args:
            safe = True
            args.remove('safe')

        if len(args) == 1:
            action_type = args[0]
            bucket = None
        elif len(args) == 2:
            action_type, bucket = args
        else:
            self._print_error("Wrong number of arguments for index command.")
            return

        if action_type not in valid_actions:
            self._print_error("\nInvalid action specified: {}\n\n"
                              "Valid actions are:\n{}".format(action_type, "\n".join(valid_actions)))
            return

        if bucket and bucket not in valid_buckets:
            self._print_error("\nInvalid bucket specified: {}\n\n"
                              "Valid buckets are:\n{}".format(bucket, "\n".join(valid_buckets)))
            return

        if safe:
            self.datastore.stop_model_validation()

        try:
            if action_type == 'reindex':
                if bucket:
                    collection = self.datastore.get_collection(bucket)
                    self.logger.info(f"Reindexing {bucket.upper()} ...")
                    collection.reindex()
                    self.logger.info("    Done!")
                else:
                    for bucket in valid_buckets:
                        collection = self.datastore.get_collection(bucket)
                        self.logger.info(f"Reindexing {bucket} ...")
                        collection.reindex()
                        self.logger.info("    Done!")
            elif action_type == 'commit':
                if bucket:
                    collection = self.datastore.get_collection(bucket)
                    collection.commit()
                    self.logger.info(f"Index {bucket.upper()} was committed.")
                else:
                    self.logger.info("Forcing commit procedure for all indexes...")
                    for bucket in valid_buckets:
                        collection = self.datastore.get_collection(bucket)
                        collection.commit()
                        self.logger.info(f"    Index {bucket.upper()} was commited.")
                    self.logger.info("All indexes committed.")
            elif action_type == 'fix_shards':
                buckets = []
                if bucket:
                    self.logger.info(f"Fixing shards on index {bucket.upper()}...")
                    buckets.append(bucket)
                else:
                    self.logger.info("Fixing shards on all indices...")
                    buckets = valid_buckets

                for bucket in buckets:
                    collection = self.datastore.get_collection(bucket)
                    collection.fix_shards()
                    self.logger.info(f"    Index {bucket.upper()} shards configuration updated.")

                self.logger.info("Completed!")
            elif action_type == 'fix_ilm':
                buckets = []
                if bucket:
                    self.logger.info(f"Fixing ILM on index {bucket.upper()}...")
                    buckets.append(bucket)
                else:
                    self.logger.info("Fixing ILM on all indices...")
                    buckets = valid_buckets

                for bucket in buckets:
                    collection = self.datastore.get_collection(bucket)
                    collection.fix_ilm()
                    self.logger.info(f"    Index {bucket.upper()} ILM configuration updated.")
            elif action_type == 'fix_replicas':
                buckets = []
                if bucket:
                    self.logger.info(f"Fixing replicas on index {bucket.upper()}...")
                    buckets.append(bucket)
                else:
                    self.logger.info("Fixing replicas on all indices...")
                    buckets = valid_buckets

                for bucket in buckets:
                    collection = self.datastore.get_collection(bucket)
                    collection.fix_replicas()
                    self.logger.info(f"    Index {bucket.upper()} replicas configuration updated.")

                self.logger.info("Completed!")
        finally:
            self.datastore.start_model_validation()

    def do_wipe(self, args):
        """
        Wipe all data from one or many buckets

        DO NOT USE ON PRODUCTION SYSTEM

        Usage:
            wipe bucket <bucket_name>
                 non_system
                 submission_data

        Actions:
            bucket           Single bucket wipe mode
            non_system       Delete all data from:
                                 alert
                                 emptyresult
                                 error
                                 file
                                 filescore
                                 result
                                 submission
                                 workflow
            submission_data  Delete all data from:
                                 emptyresult
                                 error
                                 file
                                 filescore
                                 result
                                 submission

        Parameters:
            <bucket_name>  Name of the bucket to wipe

        Examples:
            # Wipe all files
            wipe bucket file
        """
        args = self._parse_args(args)
        valid_actions = ['bucket', 'non_system', 'submission_data']
        valid_buckets = list(self.datastore.ds.get_models().keys())

        if len(args) == 1:
            action_type = args[0]
            bucket = None
        elif len(args) == 2:
            action_type, bucket = args
        else:
            self._print_error("Wrong number of arguments for wipe command.")
            return

        if action_type not in valid_actions:
            self._print_error("\nInvalid action specified: {}\n\n"
                              "Valid actions are:\n{}".format(action_type, "\n".join(valid_actions)))
            return

        if action_type == 'bucket':
            if bucket not in valid_buckets:
                self._print_error("\nInvalid bucket: {}\n\n"
                                  "Valid buckets are:\n{}".format(bucket, "\n".join(valid_buckets)))
                return

            self.datastore.get_collection(bucket).wipe()
            self.logger.info(f"Done wiping {bucket.upper()}.")
        elif action_type == 'non_system':
            non_system_buckets = [
                'alert', 'cached_file', 'emptyresult', 'error', 'file', 'filescore', 'result',
                'submission', 'submission_tree', 'submission_summary', 'workflow'
            ]
            for bucket in non_system_buckets:
                self.datastore.get_collection(bucket).wipe()
                self.logger.info(f"Done wipping {bucket.upper()}.")
        elif action_type == 'submission_data':
            submission_data_buckets = ['emptyresult', 'error', 'file', 'filescore', 'result',
                                       'submission', 'submission_tree', 'submission_summary']
            for bucket in submission_data_buckets:
                self.datastore.get_collection(bucket).wipe()
                self.logger.info(f"Done wipping {bucket.upper()}.")
        else:
            self._print_error("Invalid command parameters")

    def do_data_reset(self, args):
        """
        Completely resets the database. Does a backup of the system data, wipe every buckets then
        restores the backup.

        DO NOT USE ON PRODUCTION SYSTEM

        Usage:
            data_reset [full]

        Parameters:
            full   Does not just wipe the system bucket, also wipe all submissions and results

        Examples:
            # Reset the database
            data_reset full
        """
        args = self._parse_args(args)

        if 'full' in args:
            full = True
        else:
            full = False

        backup_file = f"/tmp/al_backup_{get_random_id()}"
        self.do_backup(backup_file)

        system_buckets = [
            'heuristic',
            'service',
            'service_delta',
            'signature',
            'user',
            'user_avatar',
            'user_favorites',
            'user_settings',
            'workflow'
        ]
        if full:
            data_buckets = [
                'alert',
                'cached_file',
                'emptyresult',
                'error',
                'file',
                'filescore',
                'result',
                'submission',
                'submission_tree',
                'submission_summary'
            ]
            system_buckets += data_buckets

        self.logger.info("\nWiping all buckets:")
        for bucket in sorted(system_buckets):
            self.datastore.get_collection(bucket).wipe()
            self.logger.info(f"    {bucket.upper()} wiped.")

        self.do_restore(backup_file)
        self.do_index("commit")
        shutil.rmtree(backup_file)
        self.logger.info("\nData reset completed.\n")

    def do_ui(self, args):
        """
        Perform UI related operations

        Usage:
            ui show_sessions [username]
            ui clear_sessions [username]

        actions:
            show_sessions      show all active sessions
            clear_sessions     Removes all active sessions

        Parameters:
            username           User use to filter sessions
                               [optional]

        Examples:
            # Clear sessions for user bob
            ui clear_sessions bob

            # Show all current sessions
            ui show_sessions
        """
        valid_func = ['clear_sessions', 'show_sessions']
        args = self._parse_args(args)

        if len(args) not in [1, 2]:
            self._print_error("Wrong number of arguments for restore command.")
            return

        func = args[0]
        if func not in valid_func:
            self._print_error(f"Invalid action '{func}' for ui command.")
            return

        if func == 'clear_sessions':
            username = None
            if len(args) == 2:
                username = args[1]

            flsk_sess = Hash(
                "flask_sessions",
                host=config.core.redis.nonpersistent.host,
                port=config.core.redis.nonpersistent.port
            )

            if not username:
                flsk_sess.delete()
                self.logger.info("All sessions where cleared.")
            else:
                for k, v in flsk_sess.items().items():
                    if v.get('username', None) == username:
                        self.logger.info(f"Removing session: {v}")
                        flsk_sess.pop(k)

                self.logger.info(f"All sessions for user '{username}' removed.")
        if func == 'show_sessions':
            username = None
            if len(args) == 2:
                username = args[1]

            flsk_sess = Hash(
                "flask_sessions",
                host=config.core.redis.nonpersistent.host,
                port=config.core.redis.nonpersistent.port
            )

            if not username:
                for k, v in flsk_sess.items().items():
                    self.logger.info(f"{v.get('username', None)} => {v}")
            else:
                self.logger.info(f'Showing sessions for user {username}:')
                for k, v in flsk_sess.items().items():
                    if v.get('username', None) == username:
                        self.logger.info(f"    {v}")


def print_banner():
    from assemblyline.common import banner
    print(banner.BANNER)


def shell_main():
    cli = ALCommandLineInterface(len(sys.argv) == 1)

    if len(sys.argv) != 1:
        cli.onecmd(" ".join([{True: f'"{x}"', False: x}[" " in x] for x in sys.argv[1:]]))
    else:
        print_banner()
        cli.cmdloop()


if __name__ == '__main__':
    try:
        shell_main()
    except KeyboardInterrupt:
        exit()




import json
import logging
import os
import shutil
import subprocess
import tempfile
import time

from copy import copy
from cart import pack_stream, unpack_stream, is_cart

from assemblyline.common import forge
from assemblyline.common.uid import get_random_id
from assemblyline.datastore.exceptions import MultiKeyError
from assemblyline.filestore import FileStoreException

config = forge.get_config()
Classification = forge.get_classification()
MAX_RETRY = 10
WORK_DIR = "/tmp/bundling"
BUNDLE_MAGIC = b'\x1f\x8b\x08'
BUNDLE_TYPE = "archive/bundle/al"

log = logging.getLogger('assemblyline.bundling')


class BundlingException(Exception):
    pass


class SubmissionNotFound(Exception):
    pass


class IncompleteBundle(Exception):
    pass


class SubmissionAlreadyExist(Exception):
    pass


# noinspection PyBroadException
def format_result(r):
    try:
        title = r['result']['sections'][0]['title_text']
        if title.startswith('Result exceeded max size.'):
            sha256 = r['response']['supplementary'][-1][1]
            with forge.get_filestore() as transport:
                oversized = json.loads(transport.get(sha256))
            oversized['oversized'] = True
            return oversized
    except Exception:  # pylint:disable=W0702
        pass

    return r


def get_results(keys, file_infos_p, storage_p):
    out = {}
    res = {}
    missing = []
    retry = 0
    while keys and retry < MAX_RETRY:
        if retry:
            time.sleep(2 ** (retry - 7))
        try:
            res.update(storage_p.get_multiple_results(keys, Classification, as_obj=False))
        except MultiKeyError as e:
            log.warning(f"Trying to get multiple results but some are missing: {str(e.keys)}")
            res.update(e.partial_output)
            missing.extend(e.keys)
        keys = [x for x in keys if x not in res and x not in missing]
        retry += 1

    results = {}
    for k, v in res.items():
        file_info = file_infos_p.get(k[:64], None)
        if file_info:
            v = format_result(v)
            if v:
                results[k] = v

    out["results"] = results
    out["missing_result_keys"] = keys + missing

    return out


def get_errors(keys, storage_p):
    out = {}
    err = {}
    missing = []
    retry = 0
    while keys and retry < MAX_RETRY:
        if retry:
            time.sleep(2 ** (retry - 7))
        try:
            err.update(storage_p.error.multiget(keys, as_obj=False))
        except MultiKeyError as e:
            log.warning(f"Trying to get multiple errors but some are missing: {str(e.keys)}")
            err.update(e.partial_output)
            missing.extend(e.keys)
        keys = [x for x in keys if x not in err and x not in missing]
        retry += 1

    out["errors"] = err
    out["missing_error_keys"] = keys + missing

    return out


def get_file_infos(keys, storage_p):
    infos = {}
    missing = []
    retry = 0
    while keys and retry < MAX_RETRY:
        if retry:
            time.sleep(2 ** (retry - 7))
        try:
            infos.update(storage_p.file.multiget(keys, as_obj=False))
        except MultiKeyError as e:
            log.warning(f"Trying to get multiple files but some are missing: {str(e.keys)}")
            infos.update(e.partial_output)
            missing.extend(e.keys)
        keys = [x for x in keys if x not in infos and x not in missing]
        retry += 1

    return infos, missing


def recursive_flatten_tree(tree):
    sha256s = []

    for key, val in tree.items():
        sha256s.extend(recursive_flatten_tree(val.get('children', {})))
        if key not in sha256s:
            sha256s.append(key)

    return list(set(sha256s))


# noinspection PyBroadException
def create_bundle(sid, working_dir=WORK_DIR):
    with forge.get_datastore() as datastore:
        temp_bundle_file = f"bundle_{get_random_id()}"
        current_working_dir = os.path.join(working_dir, temp_bundle_file)
        try:
            submission = datastore.submission.get(sid, as_obj=False)
            if submission is None:
                raise SubmissionNotFound("Can't find submission %s, skipping." % sid)
            else:
                target_file = os.path.join(working_dir, f"{temp_bundle_file}.cart")
                tgz_file = os.path.join(working_dir, f"{temp_bundle_file}.tgz")

                try:
                    os.makedirs(current_working_dir)
                except PermissionError:
                    raise
                except Exception:
                    pass

                # Create file information data
                file_tree = datastore.get_or_create_file_tree(submission,
                                                              config.submission.max_extraction_depth)['tree']
                flatten_tree = list(set(recursive_flatten_tree(file_tree) +
                                        [r[:64] for r in submission.get("results", [])]))
                file_infos, _ = get_file_infos(copy(flatten_tree), datastore)

                # Add bundling metadata
                if 'bundle.source' not in submission['metadata']:
                    submission['metadata']['bundle.source'] = config.ui.fqdn
                if Classification.enforce and 'bundle.classification' not in submission['metadata']:
                    submission['metadata']['bundle.classification'] = submission['classification']

                data = {
                    'submission': submission,
                    'files': {"list": flatten_tree, "tree": file_tree, "infos": file_infos},
                    'results': get_results(submission.get("results", []), file_infos, datastore),
                    'errors': get_errors(submission.get("errors", []), datastore)
                }

                # Save result files
                with open(os.path.join(current_working_dir, "results.json"), "w") as fp:
                    json.dump(data, fp)

                # Download all related files
                with forge.get_filestore() as filestore:
                    for sha256 in flatten_tree:
                        try:
                            filestore.download(sha256, os.path.join(current_working_dir, sha256))
                        except FileStoreException:
                            pass

                # Create the bundle
                subprocess.check_call("tar czf %s *" % tgz_file, shell=True, cwd=current_working_dir)

                with open(target_file, 'wb') as oh:
                    with open(tgz_file, 'rb') as ih:
                        pack_stream(ih, oh, {'al': {"type": BUNDLE_TYPE}, 'name': f"{sid}.tgz"})

                return target_file

        except Exception as e:
            raise BundlingException("Could not bundle submission '%s'. [%s: %s]" % (sid, type(e).__name__, str(e)))
        finally:
            if current_working_dir:
                subprocess.check_call(["rm", "-rf", current_working_dir])


# noinspection PyBroadException,PyProtectedMember
def import_bundle(path, working_dir=WORK_DIR, min_classification=Classification.UNRESTRICTED, allow_incomplete=False):
    with forge.get_datastore(archive_access=True) as datastore:
        current_working_dir = os.path.join(working_dir, get_random_id())
        res_file = os.path.join(current_working_dir, "results.json")
        try:
            os.makedirs(current_working_dir)
        except Exception:
            pass

        with open(path, 'rb') as original_file:
            if is_cart(original_file.read(256)):
                original_file.seek(0)

                extracted_fd, extracted_path = tempfile.mkstemp()
                extracted_file = os.fdopen(extracted_fd, 'wb')

                try:
                    hdr, _ = unpack_stream(original_file, extracted_file)
                    if hdr.get('al', {}).get('type', 'unknown') != BUNDLE_TYPE:
                        raise BundlingException(f"Not a valid CaRTed bundle, should be of type: {BUNDLE_TYPE}")
                finally:
                    extracted_file.close()
            else:
                extracted_path = path

        # Extract  the bundle
        try:
            subprocess.check_call(["tar", "-zxf", extracted_path, "-C", current_working_dir])
        except subprocess.CalledProcessError:
            raise BundlingException("Bundle decompression failed. Not a valid bundle...")

        with open(res_file, 'rb') as fh:
            data = json.load(fh)

        submission = data['submission']
        results = data['results']
        files = data['files']
        errors = data['errors']

        try:
            sid = submission['sid']
            # Check if we have all the service results
            for res_key in submission['results']:
                if res_key not in results['results'].keys() and not allow_incomplete:
                    raise IncompleteBundle("Incomplete results in bundle. Skipping %s..." % sid)

            # Check if we have all files
            for sha256 in list(set([x[:64] for x in submission['results']])):
                if sha256 not in files['infos'].keys() and not allow_incomplete:
                    raise IncompleteBundle("Incomplete files in bundle. Skipping %s..." % sid)

            # Check if we all errors
            for err_key in submission['errors']:
                if err_key not in errors['errors'].keys() and not allow_incomplete:
                    raise IncompleteBundle("Incomplete errors in bundle. Skipping %s..." % sid)

            if datastore.submission.get(sid, as_obj=False):
                raise SubmissionAlreadyExist("Submission %s already exists." % sid)

            # Make sure bundle's submission meets minimum classification and save the submission
            submission['classification'] = Classification.max_classification(submission['classification'],
                                                                             min_classification)
            submission.update(Classification.get_access_control_parts(submission['classification']))
            datastore.submission.save(sid, submission)

            # Make sure files meet minimum classification and save the files
            with forge.get_filestore() as filestore:
                for f, f_data in files['infos'].items():
                    f_classification = Classification.max_classification(f_data['classification'], min_classification)
                    datastore.save_or_freshen_file(f, f_data, f_data['expiry_ts'], f_classification,
                                                   cl_engine=Classification)
                    try:
                        filestore.upload(os.path.join(current_working_dir, f), f)
                    except IOError:
                        pass

            # Make sure results meet minimum classification and save the results
            for key, res in results['results'].items():
                if key.endswith(".e"):
                    datastore.emptyresult.save(key, {"expiry_ts": res['expiry_ts']})
                else:
                    res['classification'] = Classification.max_classification(res['classification'], min_classification)
                    datastore.result.save(key, res)

            # Make sure errors meet minimum classification and save the errors
            for ekey, err in errors['errors'].items():
                datastore.error.save(ekey, err)

            return submission
        finally:
            try:
                os.remove(extracted_path)
            except Exception:
                pass

            try:
                os.remove(path)
            except Exception:
                pass

            try:
                shutil.rmtree(current_working_dir, ignore_errors=True)
            except Exception:
                pass



import json
import redis

from assemblyline.common.forge import get_pubsub_redis
from assemblyline.remote.datatypes import retry_call, log, decode


class CommsQueue:
    def __init__(self, names, host=None, port=None, private=False):
        self.c = get_pubsub_redis(host, port, private)
        self.p = retry_call(self.c.pubsub)
        if not isinstance(names, list):
            names = [names]
        self.names = names
        self._connected = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        retry_call(self.p.unsubscribe)

    def _connect(self):
        if not self._connected:
            retry_call(self.p.subscribe, self.names)
            self._connected = True

    def close(self):
        retry_call(self.p.close)

    def listen(self, blocking=True):
        retried = False
        while True:
            self._connect()
            try:
                if blocking:
                    i = self.p.listen()
                    v = next(i)
                else:
                    v = self.p.get_message()
                    if v is None:
                        yield None
                        continue

                if isinstance(v, dict) and v.get('type', None) == 'message':
                    data = decode(v.get('data', 'null'))
                    yield data
            except redis.ConnectionError:
                log.warning('No connection to Redis, reconnecting...')
                self._connected = False
                retried = True
            finally:
                if self._connected and retried:
                    log.info('Reconnected to Redis!')
                    retried = False

    def publish(self, message):
        for name in self.names:
            retry_call(self.c.publish, name, json.dumps(message))




import concurrent.futures
import json
import time

from typing import Union, List
from datetime import datetime

import elasticapm
import elasticsearch
from assemblyline.datastore.exceptions import MultiKeyError, VersionConflictException

from assemblyline.common import forge
from assemblyline.common.dict_utils import recursive_update, flatten
from assemblyline.common.isotime import now_as_iso
from assemblyline.datastore import Collection, log
from assemblyline.odm import Model, DATEFORMAT
from assemblyline.odm.models.alert import Alert
from assemblyline.odm.models.cached_file import CachedFile
from assemblyline.odm.models.emptyresult import EmptyResult
from assemblyline.odm.models.error import Error
from assemblyline.odm.models.file import File
from assemblyline.odm.models.filescore import FileScore
from assemblyline.odm.models.heuristic import Heuristic
from assemblyline.odm.models.result import Result
from assemblyline.odm.models.service import Service
from assemblyline.odm.models.service_delta import ServiceDelta
from assemblyline.odm.models.signature import Signature
from assemblyline.odm.models.submission import Submission
from assemblyline.odm.models.submission_summary import SubmissionSummary
from assemblyline.odm.models.submission_tree import SubmissionTree
from assemblyline.odm.models.user import User
from assemblyline.odm.models.user_favorites import UserFavorites
from assemblyline.odm.models.user_settings import UserSettings
from assemblyline.odm.models.safelist import Safelist
from assemblyline.odm.models.workflow import Workflow

config = forge.get_config()


class AssemblylineDatastore(object):
    def __init__(self, datastore_object):

        self.ds = datastore_object
        self.ds.register('alert', Alert)
        self.ds.register('cached_file', CachedFile)
        self.ds.register('emptyresult', EmptyResult)
        self.ds.register('error', Error)
        self.ds.register('file', File)
        self.ds.register('filescore', FileScore)
        self.ds.register('heuristic', Heuristic)
        self.ds.register('result', Result)
        self.ds.register('service', Service)
        self.ds.register('service_delta', ServiceDelta)
        self.ds.register('signature', Signature)
        self.ds.register('submission', Submission)
        self.ds.register('submission_tree', SubmissionTree)
        self.ds.register('submission_summary', SubmissionSummary)
        self.ds.register('user', User)
        self.ds.register('user_avatar')
        self.ds.register('user_favorites', UserFavorites)
        self.ds.register('user_settings', UserSettings)
        self.ds.register('safelist', Safelist)
        self.ds.register('workflow', Workflow)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ds.close()

    def stop_model_validation(self):
        self.ds.validate = False

    def start_model_validation(self):
        self.ds.validate = True

    def enable_archive_access(self):
        self.ds.archive_access = True

    def disable_archive_access(self):
        self.ds.archive_access = False

    @property
    def alert(self) -> Collection[Alert]:
        return self.ds.alert

    @property
    def cached_file(self) -> Collection[CachedFile]:
        return self.ds.cached_file

    @property
    def emptyresult(self) -> Collection[EmptyResult]:
        return self.ds.emptyresult

    @property
    def error(self) -> Collection[Error]:
        return self.ds.error

    @property
    def file(self) -> Collection[File]:
        return self.ds.file

    @property
    def filescore(self) -> Collection[FileScore]:
        return self.ds.filescore

    @property
    def heuristic(self) -> Collection[Heuristic]:
        return self.ds.heuristic

    @property
    def result(self) -> Collection[Result]:
        return self.ds.result

    @property
    def service(self) -> Collection[Service]:
        return self.ds.service

    @property
    def service_client(self) -> Collection:
        return self.ds.service_client

    @property
    def service_delta(self) -> Collection[ServiceDelta]:
        return self.ds.service_delta

    @property
    def signature(self) -> Collection[Signature]:
        return self.ds.signature

    @property
    def submission(self) -> Collection[Submission]:
        return self.ds.submission

    @property
    def submission_summary(self) -> Collection[SubmissionSummary]:
        return self.ds.submission_summary

    @property
    def submission_tree(self) -> Collection[SubmissionTree]:
        return self.ds.submission_tree

    @property
    def user(self) -> Collection[User]:
        return self.ds.user

    @property
    def user_avatar(self) -> Collection:
        return self.ds.user_avatar

    @property
    def user_favorites(self) -> Collection[UserFavorites]:
        return self.ds.user_favorites

    @property
    def user_settings(self) -> Collection[UserSettings]:
        return self.ds.user_settings

    @property
    def vm(self) -> Collection:
        return self.ds.vm

    @property
    def safelist(self) -> Collection[Safelist]:
        return self.ds.safelist

    @property
    def workflow(self) -> Collection[Workflow]:
        return self.ds.workflow

    def get_collection(self, collection_name: str) -> Collection:
        if collection_name in self.ds.get_models():
            return getattr(self, collection_name)
        else:
            raise AttributeError(f'Collection {collection_name} does not exist.')

    @staticmethod
    def create_empty_result_from_key(key, cl_engine=forge.get_classification(), as_obj=True):
        sha256, svc_name, svc_version, _ = key.split(".", 3)
        svc_version = svc_version[1:]

        data = Result({
            "archive_ts": now_as_iso(config.datastore.ilm.days_until_archive * 24 * 60 * 60),
            "expiry_ts": now_as_iso(config.datastore.ilm.days_until_archive * 24 * 60 * 60),
            "classification": cl_engine.UNRESTRICTED,
            "response": {
                "service_name": svc_name,
                "service_version": svc_version,
            },
            "sha256": sha256
        })
        if as_obj:
            return data
        else:
            return data.as_primitives()

    @elasticapm.capture_span(span_type='datastore')
    def multi_index_bulk(self, bulk_plans):
        max_retry_backoff = 10
        retries = 0
        while True:
            try:
                plan = "\n".join([p.get_plan_data() for p in bulk_plans])
                ret_val = self.ds.client.bulk(body=plan)
                return ret_val
            except (elasticsearch.exceptions.ConnectionError,
                    elasticsearch.exceptions.ConnectionTimeout,
                    elasticsearch.exceptions.AuthenticationException):
                log.warning(f"No connection to Elasticsearch server(s): "
                            f"{' | '.join(self.ds.get_hosts(safe=True))}"
                            f", retrying...")
                time.sleep(min(retries, max_retry_backoff))
                self.ds.connection_reset()
                retries += 1

            except elasticsearch.exceptions.TransportError as e:
                err_code, msg, cause = e.args
                if err_code == 503 or err_code == '503':
                    log.warning("Looks like index is not ready yet, retrying...")
                    time.sleep(min(retries, max_retry_backoff))
                    self.ds.connection_reset()
                    retries += 1
                elif err_code == 429 or err_code == '429':
                    log.warning("Elasticsearch is too busy to perform the requested task, "
                                "we will wait a bit and retry...")
                    time.sleep(min(retries, max_retry_backoff))
                    self.ds.connection_reset()
                    retries += 1

                else:
                    raise

    @elasticapm.capture_span(span_type='datastore')
    def delete_submission_tree_bulk(self, sid, cl_engine=forge.get_classification(), cleanup=True, transport=None):
        submission = self.submission.get(sid, as_obj=False)
        if not submission:
            return

        # Create plans
        s_plan = self.submission.get_bulk_plan()
        st_plan = self.submission_tree.get_bulk_plan()
        ss_plan = self.submission_summary.get_bulk_plan()
        e_plan = self.error.get_bulk_plan()
        er_plan = self.emptyresult.get_bulk_plan()
        r_plan = self.result.get_bulk_plan()
        f_plan = self.file.get_bulk_plan()

        # Add delete operation for submission and cache
        s_plan.add_delete_operation(sid)
        for x in self.submission_tree.stream_search(f"id:{sid}*", fl="id,_index", as_obj=False):
            st_plan.add_delete_operation(x['id'], index=x['_index'])
        for x in self.submission_summary.stream_search(f"id:{sid}*", fl="id,_index", as_obj=False):
            ss_plan.add_delete_operation(x['id'], index=x['_index'])

        # Gather file list
        errors = submission['errors']
        results = submission["results"]
        files = set()
        fix_classification_files = set()
        supp_map = {}

        temp_files = [x[:64] for x in errors]
        temp_files.extend([x[:64] for x in results])
        temp_files = set(temp_files)

        # Inspect each files to see if they are reused
        for temp in temp_files:
            # Hunt for supplementary files
            supp_list = set()
            for res in self.result.stream_search(f"id:{temp}* AND response.supplementary.sha256:*",
                                                 fl="id", as_obj=False):
                if res['id'] in results:
                    result = self.result.get(res['id'], as_obj=False)
                    for supp in result['response']['supplementary']:
                        supp_list.add(supp['sha256'])

            # Check if we delete or update classification
            if self.submission.search(f"errors:{temp}* OR results:{temp}*", rows=0, as_obj=False)["total"] < 2:
                files.add(temp)
                files = files.union(supp_list)
            else:
                fix_classification_files.add(temp)
                supp_map[temp] = supp_list

        # Filter results and errors
        errors = [x for x in errors if x[:64] in files]
        results = [x for x in results if x[:64] in files]

        # Delete files, errors, results that were only used once
        for e in errors:
            e_plan.add_delete_operation(e)
        for r in results:
            if r.endswith(".e"):
                er_plan.add_delete_operation(r)
            else:
                r_plan.add_delete_operation(r)
        for f in files:
            f_plan.add_delete_operation(f)
            if transport:
                transport.delete(f)

        if fix_classification_files and cleanup:
            # Fix classification for the files that remain in the system
            for f in fix_classification_files:
                cur_file = self.file.get(f, as_obj=False)
                if cur_file:
                    # Find possible classification for the file in the system
                    query = f"NOT id:{sid} AND (files.sha256:{f} OR results:{f}* OR errors:{f}*)"
                    classifications = list(self.submission.facet('classification', query=query).keys())

                    if len(classifications) > 0:
                        new_file_class = classifications[0]
                    else:
                        new_file_class = cl_engine.UNRESTRICTED

                    for c in classifications:
                        new_file_class = cl_engine.min_classification(new_file_class, c)

                    # Find the results for that classification and alter them if the new classification does not match
                    for item in self.result.stream_search(f"id:{f}*", fl="classification,id,_index", as_obj=False):
                        new_class = cl_engine.max_classification(
                            item.get('classification', cl_engine.UNRESTRICTED), new_file_class)
                        if item.get('classification', cl_engine.UNRESTRICTED) != new_class:
                            data = cl_engine.get_access_control_parts(new_class)
                            data['classification'] = new_class
                            r_plan.add_update_operation(item['id'], data, index=item['_index'])

                    # Alter the file classification if the new classification does not match
                    if cur_file['classification'] != new_file_class:
                        data = cl_engine.get_access_control_parts(new_file_class)
                        data['classification'] = new_file_class
                        f_plan.add_update_operation(f, data)
                    # Fix associated supplementary files
                    for supp in supp_map.get(f, set()):
                        cur_supp = self.file.get(supp, as_obj=False)
                        if cur_supp:
                            if cur_supp['classification'] != new_file_class:
                                data = cl_engine.get_access_control_parts(new_file_class)
                                data['classification'] = new_file_class
                                f_plan.add_update_operation(supp, data)

        # Proceed with plan
        self.multi_index_bulk([s_plan, st_plan, ss_plan, e_plan, er_plan, r_plan, f_plan])

    @elasticapm.capture_span(span_type='datastore')
    def delete_submission_tree(self, sid, cl_engine=forge.get_classification(), cleanup=True, transport=None):
        submission = self.submission.get(sid, as_obj=False)
        if not submission:
            return

        # Gather file list
        errors = submission['errors']
        results = submission["results"]
        files = set()
        fix_classification_files = set()
        supp_map = {}

        temp_files = [x[:64] for x in errors]
        temp_files.extend([x[:64] for x in results])
        temp_files = set(temp_files)

        # Inspect each files to see if they are reused
        for temp in temp_files:
            # Hunt for supplementary files
            supp_list = set()
            for res in self.result.stream_search(f"id:{temp}* AND response.supplementary.sha256:*",
                                                 fl="id", as_obj=False):
                if res['id'] in results:
                    result = self.result.get(res['id'], as_obj=False)
                    for supp in result['response']['supplementary']:
                        supp_list.add(supp['sha256'])

            # Check if we delete or update classification
            if self.submission.search(f"errors:{temp}* OR results:{temp}*", rows=0, as_obj=False)["total"] < 2:
                files.add(temp)
                files = files.union(supp_list)
            else:
                fix_classification_files.add(temp)
                supp_map[temp] = supp_list

        # Filter results and errors
        errors = [x for x in errors if x[:64] in files]
        results = [x for x in results if x[:64] in files]

        # Delete files, errors, results that were only used once
        for e in errors:
            self.error.delete(e)
        for r in results:
            if r.endswith(".e"):
                self.emptyresult.delete(r)
            else:
                self.result.delete(r)
        for f in files:
            self.file.delete(f)
            if transport:
                transport.delete(f)

        if fix_classification_files and cleanup:
            # Fix classification for the files that remain in the system
            for f in fix_classification_files:
                cur_file = self.file.get(f, as_obj=False)
                if cur_file:
                    # Find possible classification for the file in the system
                    query = f"NOT id:{sid} AND (files.sha256:{f} OR results:{f}* OR errors:{f}*)"
                    classifications = list(self.submission.facet('classification', query=query).keys())

                    if len(classifications) > 0:
                        new_file_class = classifications[0]
                    else:
                        new_file_class = cl_engine.UNRESTRICTED

                    for c in classifications:
                        new_file_class = cl_engine.min_classification(new_file_class, c)

                    # Find the results for that classification and alter them if the new classification does not match
                    for item in self.result.stream_search(f"id:{f}*", fl="classification,id", as_obj=False):
                        new_class = cl_engine.max_classification(
                            item.get('classification', cl_engine.UNRESTRICTED), new_file_class)
                        if item.get('classification', cl_engine.UNRESTRICTED) != new_class:
                            parts = cl_engine.get_access_control_parts(new_class)
                            update_params = [(Collection.UPDATE_SET, 'classification', new_class)]
                            update_params.extend([(Collection.UPDATE_SET, k, v) for k, v in parts.items()])
                            self.result.update(item['id'], update_params)

                    # Alter the file classification if the new classification does not match
                    if cur_file['classification'] != new_file_class:
                        parts = cl_engine.get_access_control_parts(new_file_class)
                        update_params = [(Collection.UPDATE_SET, 'classification', new_file_class)]
                        update_params.extend([(Collection.UPDATE_SET, k, v) for k, v in parts.items()])
                        self.file.update(f, update_params)

                    # Fix associated supplementary files
                    for supp in supp_map.get(f, set()):
                        cur_supp = self.file.get(supp, as_obj=False)
                        if cur_supp:
                            if cur_supp['classification'] != new_file_class:
                                parts = cl_engine.get_access_control_parts(new_file_class)
                                update_params = [(Collection.UPDATE_SET, 'classification', new_file_class)]
                                update_params.extend([(Collection.UPDATE_SET, k, v) for k, v in parts.items()])
                                self.file.update(supp, update_params)

        # Delete the submission and cached trees and summaries
        self.submission.delete(sid)
        for t in [x['id'] for x in self.submission_tree.stream_search(f"id:{sid}*", fl="id", as_obj=False)]:
            self.submission_tree.delete(t)
        for s in [x['id'] for x in self.submission_summary.stream_search(f"id:{sid}*", fl="id", as_obj=False)]:
            self.submission_summary.delete(s)

    @elasticapm.capture_span(span_type='datastore')
    def get_all_heuristics(self):
        return {h['id']: h for h in self.ds.heuristic.stream_search("id:*", as_obj=False)}

    @elasticapm.capture_span(span_type='datastore')
    def get_multiple_results(self, keys, cl_engine=forge.get_classification(), as_obj=False):
        results = {k: self.create_empty_result_from_key(k, cl_engine, as_obj=as_obj)
                   for k in keys if k.endswith(".e")}
        keys = [k for k in keys if not k.endswith(".e")]
        try:
            results.update(self.result.multiget(keys, as_dictionary=True, as_obj=as_obj))
        except MultiKeyError as e:
            log.warning(f"Trying to get multiple results but some are missing: {str(e.keys)}")
            results.update(e.partial_output)
        return results

    @elasticapm.capture_span(span_type='datastore')
    def get_single_result(self, key, cl_engine=forge.get_classification(), as_obj=False):
        if key.endswith(".e"):
            data = self.create_empty_result_from_key(key, cl_engine, as_obj=as_obj)
        else:
            data = self.result.get(key, as_obj=False)

        return data

    @elasticapm.capture_span(span_type='datastore')
    def get_file_submission_meta(self, sha256, fields, access_control=None):
        query = f"files.sha256:{sha256} OR results:{sha256}*"
        with concurrent.futures.ThreadPoolExecutor(len(fields)) as executor:
            res = {field: executor.submit(self.submission.facet,
                                          field,
                                          query=query,
                                          limit=100,
                                          access_control=access_control)
                   for field in fields}

        return {k.split(".")[-1]: v.result() for k, v in res.items()}

    @elasticapm.capture_span(span_type='datastore')
    def get_file_list_from_keys(self, keys):
        # TODO: needed?
        if len(keys) == 0:
            return {}
        keys = [x for x in list(keys) if not x.endswith(".e")]
        items = self.result.multiget(keys)

        out = {}
        for key, item in items.items():
            extracted = item['response']['extracted']
            if len(extracted) == 0:
                continue
            if key[:64] not in out:
                out[key[:64]] = []
            out[key[:64]].extend(extracted)

        return out

    @elasticapm.capture_span(span_type='datastore')
    def get_file_scores_from_keys(self, keys):
        # TODO: needed?
        if len(keys) == 0:
            return {}
        keys = [x for x in list(keys) if not x.endswith(".e")]
        items = self.result.multiget(keys)

        scores = {x[:64]: 0 for x in keys}
        for key, item in items.items():
            score = item["result"]["score"]
            scores[key[:64]] += score

        return scores

    @elasticapm.capture_span(span_type='datastore')
    def get_signature_last_modified(self, sig_type=None):
        if sig_type is None:
            sig_type = "*"

        res = self.signature.search(f"type:{sig_type}", fl="last_modified",
                                    sort="last_modified desc", rows=1, as_obj=False)
        if res['total'] > 0:
            return res['items'][0]['last_modified']
        return '1970-01-01T00:00:00.000000Z'

    @elasticapm.capture_span(span_type='datastore')
    def get_or_create_file_tree(self, submission, max_depth, cl_engine=forge.get_classification(),
                                user_classification=None):
        if user_classification is not None:
            user_classification = cl_engine.normalize_classification(user_classification, long_format=False)
            cache_key = f"{submission['sid']}_{user_classification}"
            for illegal_char in [" ", ":", "/"]:
                cache_key = cache_key.replace(illegal_char, "")
        else:
            cache_key = submission['sid']

        if isinstance(submission, Model):
            submission = submission.as_primitives()

        num_files = len(list(set([x[:64] for x in submission['results']])))
        max_score = submission['max_score']

        cached_tree = self.submission_tree.get_if_exists(cache_key, as_obj=False)
        if cached_tree:
            tree = json.loads(cached_tree['tree'])
            if self._is_valid_tree(tree, num_files, max_score):
                return {
                    "tree": tree,
                    "classification": cached_tree['classification'],
                    "filtered": cached_tree['filtered'],
                    "partial": False
                }

        partial = submission['state'] != 'completed'
        files = {}
        scores = {}
        missing_files = []
        file_hashes = [x[:64] for x in submission['results']]
        file_hashes.extend([x[:64] for x in submission['errors']])
        file_hashes.extend([f['sha256'] for f in submission['files']])
        try:
            temp_file_data_map = self.file.multiget(list(set(file_hashes)), as_dictionary=True, as_obj=False)
        except MultiKeyError as e:
            log.warning(f"Trying to generate file tree but we are missing file(s): {str(e.keys)}")
            temp_file_data_map = e.partial_output
            missing_files = e.keys
            partial = True
        forbidden_files = set()

        max_classification = cl_engine.UNRESTRICTED
        file_data_map = {}
        for key, value in temp_file_data_map.items():
            if user_classification and not cl_engine.is_accessible(user_classification, value['classification']):
                partial = True
                forbidden_files.add(key)
                continue
            file_data_map[key] = value
            max_classification = cl_engine.max_classification(max_classification, value['classification'])

        try:
            results_data = self.result.multiget([x for x in submission['results'] if not x.endswith(".e")],
                                                as_obj=False)
        except MultiKeyError as e:
            log.warning(f"Trying to generate file tree but we are missing result(s): {str(e.keys)}")
            results_data = e.partial_output
            partial = True

        for key, item in results_data.items():
            sha256 = key[:64]

            # Get scores
            if sha256 not in scores:
                scores[sha256] = 0
            scores[sha256] += item["result"]["score"]

            # Get files
            extracted = item['response']['extracted']
            if len(extracted) == 0:
                continue
            if sha256 not in files:
                files[sha256] = []
            files[sha256].extend(extracted)

        tree_cache = []

        def recurse_tree(child_p, placeholder, parents_p, lvl=0):
            if lvl == max_depth + 1:
                # Enforce depth protection while building the tree
                return

            c_sha256 = child_p['sha256']
            c_name = child_p['name']
            if c_sha256 in placeholder:
                placeholder[c_sha256]['name'].append(c_name)
            else:
                children_list = {}
                truncated = False
                child_list = files.get(c_sha256, [])
                for new_child in child_list:
                    if new_child['sha256'] in tree_cache:
                        truncated = True
                        continue
                    tree_cache.append(child['sha256'])

                    if new_child['sha256'] not in parents_p:
                        recurse_tree(new_child, children_list,
                                     parents_p + [c_sha256], lvl + 1)

                try:
                    placeholder[c_sha256] = {
                        "name": [c_name],
                        "type": file_data_map[c_sha256]['type'],
                        "sha256": file_data_map[c_sha256]['sha256'],
                        "size": file_data_map[c_sha256]['size'],
                        "children": children_list,
                        "truncated": truncated,
                        "score": scores.get(c_sha256, 0),
                    }
                except KeyError:
                    if c_sha256 not in forbidden_files and c_sha256 not in missing_files:
                        file_data_map[c_sha256] = self.file.get(c_sha256, as_obj=False)
                        placeholder[c_sha256] = {
                            "name": [c_name],
                            "type": file_data_map[c_sha256]['type'],
                            "sha256": file_data_map[c_sha256]['sha256'],
                            "size": file_data_map[c_sha256]['size'],
                            "children": children_list,
                            "truncated": truncated,
                            "score": scores.get(c_sha256, 0),
                        }

        tree = {}
        for f in submission['files']:
            sha256 = f['sha256']
            name = f['name']

            if sha256 in tree:
                tree[sha256]['name'].append(name)
            else:
                parents = [sha256]
                children = {}
                c_list = files.get(sha256, [])
                for child in c_list:
                    tree_cache.append(child['sha256'])
                    recurse_tree(child, children, parents)

                try:
                    tree[sha256] = {
                        "name": [name],
                        "children": children,
                        "type": file_data_map[sha256]['type'],
                        "sha256": file_data_map[sha256]['sha256'],
                        "size": file_data_map[sha256]['size'],
                        "truncated": False,
                        "score": scores.get(sha256, 0),
                    }
                except KeyError:
                    if sha256 not in forbidden_files and sha256 not in missing_files:
                        file_data_map[sha256] = self.file.get(sha256, as_obj=False)
                        tree[sha256] = {
                            "name": [name],
                            "children": children,
                            "type": file_data_map[sha256]['type'],
                            "sha256": file_data_map[sha256]['sha256'],
                            "size": file_data_map[sha256]['size'],
                            "truncated": False,
                            "score": scores.get(sha256, 0),
                        }

        if not partial:
            cached_tree = {
                'expiry_ts': now_as_iso(config.datastore.ilm.days_until_archive * 24 * 60 * 60),
                'tree': json.dumps(tree),
                'classification': max_classification,
                'filtered': len(forbidden_files) > 0
            }

            self.submission_tree.save(cache_key, cached_tree)

        return {
            'tree': tree,
            'classification': max_classification,
            'filtered': len(forbidden_files) > 0,
            'partial': partial
        }

    @staticmethod
    def _is_valid_tree(tree, num_files, max_score):
        def _count_children(sub_tree, cur_files, cur_score):
            temp_score = cur_score
            for k, v in sub_tree.items():
                if v['score'] > temp_score:
                    temp_score = v['score']
                cur_files.append(k)
                cur_files, temp_score = _count_children(v.get("children", {}), cur_files, temp_score)
            return cur_files, temp_score

        files, tree_score = _count_children(tree, [], 0)
        files = list(set(files))

        if len(files) < num_files:
            return False

        if tree_score != max_score:
            return False

        return True

    @elasticapm.capture_span(span_type='datastore')
    def get_summary_from_keys(self, keys, cl_engine=forge.get_classification(), user_classification=None):
        out = {
            "tags": [],
            "attack_matrix": [],
            "heuristics": {
                "info": [],
                "safe": [],
                "suspicious": [],
                "malicious": []
            },
            "classification": cl_engine.UNRESTRICTED,
            "filtered": False
        }
        done_map = {
            "heuristics": set(),
            "attack": set(),
            "tags": set()
        }

        if len(keys) == 0:
            return out

        keys = [x for x in list(keys) if not x.endswith(".e")]
        file_keys = list(set([x[:64] for x in keys]))
        try:
            items = self.result.multiget(keys, as_obj=False)
        except MultiKeyError as e:
            # Generate partial summaries even if results are missing
            log.warning(f"Trying to generate summary but we are missing result(s): {str(e.keys)}")
            items = e.partial_output
            out['missing_results'] = e.keys
        try:
            files = self.file.multiget(file_keys, as_obj=False)
        except MultiKeyError as e:
            # Generate partial summaries even if results are missing
            log.warning(f"Trying to generate summary but we are missing file(s): {str(e.keys)}")
            files = e.partial_output
            out['missing_files'] = e.keys

        for key, item in items.items():
            for section in item.get('result', {}).get('sections', []):
                file_classification = files.get(key[:64], {}).get('classification', section['classification'])
                if user_classification:
                    if not cl_engine.is_accessible(user_classification, section['classification']):
                        out["filtered"] = True
                        continue
                    if not cl_engine.is_accessible(user_classification, file_classification):
                        out["filtered"] = True
                        continue

                out["classification"] = cl_engine.max_classification(out["classification"], section['classification'])
                out["classification"] = cl_engine.max_classification(out["classification"], file_classification)

                h_type = "info"

                if section.get('heuristic', False):
                    # Get the heuristics data
                    if section['heuristic']['score'] < 0:
                        h_type = "safe"
                    elif section['heuristic']['score'] < 300:
                        h_type = "info"
                    elif section['heuristic']['score'] < 1000:
                        h_type = "suspicious"
                    else:
                        h_type = "malicious"

                    cache_key = f"{section['heuristic']['heur_id']}_{key}"
                    if cache_key not in done_map['heuristics']:
                        out['heuristics'][h_type].append({
                            'heur_id': section['heuristic']['heur_id'],
                            'name': section['heuristic']['name'],
                            'key': key
                        })
                        done_map['heuristics'].add(cache_key)

                    for attack in section['heuristic'].get('attack', []):
                        # Get attack matrix data
                        attack_id = attack['attack_id']

                        cache_key = f"{attack_id}_{key}"
                        if cache_key not in done_map['attack']:
                            out['attack_matrix'].append({
                                "key": key,
                                "attack_id": attack_id,
                                "h_type": h_type,
                                "name": attack['pattern'],
                                "categories": attack['categories']
                            })
                            done_map['attack'].add(cache_key)

                # Get tagging data
                for tag_type, tags in flatten(section.get('tags', {})).items():
                    if tags is not None:
                        for tag in tags:
                            cache_key = f"{tag_type}_{tag}_{key}"

                            if cache_key not in done_map['tags']:
                                out['tags'].append({
                                    'type': tag_type,
                                    'h_type': h_type,
                                    'short_type': tag_type.rsplit(".", 1)[-1],
                                    'value': tag,
                                    'key': key,
                                    'safelisted': False
                                })
                                done_map['tags'].add(cache_key)

                # Get safelisted tag data
                for tag_type, tags in section.get('safelisted_tags', {}).items():
                    if tags is not None:
                        for tag in tags:
                            cache_key = f"{tag_type}_{tag}_{key}"

                            if cache_key not in done_map['tags']:
                                out['tags'].append({
                                    'type': tag_type,
                                    'h_type': h_type,
                                    'short_type': tag_type.rsplit(".", 1)[-1],
                                    'value': tag,
                                    'key': key,
                                    'safelisted': True
                                })
                                done_map['tags'].add(cache_key)

        return out

    @elasticapm.capture_span(span_type='datastore')
    def get_tag_list_from_keys(self, keys):
        if len(keys) == 0:
            return []
        keys = [x for x in list(keys) if not x.endswith(".e")]
        items = self.result.multiget(keys, as_obj=False)

        out = []
        for key, item in items.items():
            for section in item.get('result', {}).get('sections', []):
                for tag_type, tags in flatten(section.get('tags', {})).items():
                    if tags is not None:
                        for tag in tags:
                            out.append({
                                'type': tag_type,
                                'short_type': tag_type.rsplit(".", 1)[-1],
                                'value': tag,
                                'key': key,
                                'safelisted': False
                            })
                for tag_type, tags in section.get('safelisted_tags', {}).items():
                    if tags is not None:
                        for tag in tags:
                            out.append({
                                'type': tag_type,
                                'short_type': tag_type.rsplit(".", 1)[-1],
                                'value': tag,
                                'key': key,
                                'safelisted': True
                            })

        return out

    @elasticapm.capture_span(span_type='datastore')
    def get_attack_matrix_from_keys(self, keys):
        if len(keys) == 0:
            return []
        keys = [x for x in list(keys) if not x.endswith(".e")]
        items = self.result.multiget(keys, as_obj=False)

        out = []
        for key, item in items.items():
            for section in item.get('result', {}).get('sections', []):
                attack_id = section.get('heuristic', {}).get('attack_id', None)
                if attack_id:
                    out.append({
                        "key": key,
                        "attack_id": attack_id,
                        "name": section['heuristic']['attack_pattern'],
                        "categories": section['heuristic']['attack_categories']
                    })

        return out

    @elasticapm.capture_span(span_type='datastore')
    def get_service_with_delta(self, service_name, version=None, as_obj=True) -> Union[Service, dict, None]:
        svc = self.ds.service_delta.get(service_name)
        if svc is None:
            return None

        if version is not None:
            svc.version = version

        svc_version_data = self.ds.service.get(f"{service_name}_{svc.version}")
        if svc_version_data is None:
            return None

        svc_version_data = recursive_update(svc_version_data.as_primitives(strip_null=True),
                                            svc.as_primitives(strip_null=True),
                                            stop_keys=['config'])
        if as_obj:
            return Service(svc_version_data)
        else:
            return svc_version_data

    @elasticapm.capture_span(span_type='datastore')
    def get_stat_for_heuristic(self, p_id):
        query = f"result.sections.heuristic.heur_id:{p_id}"
        stats = self.ds.result.stats("result.score", query=query)

        if stats['count'] == 0:
            up_stats = {'count': 0, 'min': 0, 'max': 0, 'avg': 0, 'sum': 0}
        else:
            first = self.ds.result.search(query=query, fl='created', rows=1,
                                          sort="created asc", as_obj=False, use_archive=True)['items'][0]['created']
            last = self.ds.result.search(query=query, fl='created', rows=1,
                                         sort="created desc", as_obj=False, use_archive=True)['items'][0]['created']
            up_stats = {
                'count': stats['count'],
                'min': int(stats['min']),
                'max': int(stats['max']),
                'avg': int(stats['avg']),
                'sum': int(stats['sum']),
                'first_hit': first,
                'last_hit': last
            }

        self.ds.heuristic.update(p_id, [
            (self.ds.heuristic.UPDATE_SET, 'stats', up_stats)
        ])

        return up_stats

    @elasticapm.capture_span(span_type='datastore')
    def calculate_heuristic_stats(self):
        heur_list = sorted([(x['heur_id'])
                            for x in self.ds.heuristic.stream_search("heur_id:*", fl="heur_id,name,classification",
                                                                     as_obj=False)])

        with concurrent.futures.ThreadPoolExecutor(max(min(len(heur_list), 20), 1)) as executor:
            for heur_id in heur_list:
                executor.submit(self.get_stat_for_heuristic, heur_id)

    @elasticapm.capture_span(span_type='datastore')
    def get_stat_for_signature(self, p_id, p_source, p_name, p_type):
        query = f'result.sections.tags.file.rule.{p_type}:"{p_source}.{p_name}"'
        stats = self.ds.result.stats("result.score", query=query)
        if stats['count'] == 0:
            up_stats = {'count': 0, 'min': 0, 'max': 0, 'avg': 0, 'sum': 0}
        else:
            first = self.ds.result.search(query=query, fl='created', rows=1,
                                          sort="created asc", as_obj=False, use_archive=True)['items'][0]['created']
            last = self.ds.result.search(query=query, fl='created', rows=1,
                                         sort="created desc", as_obj=False, use_archive=True)['items'][0]['created']
            up_stats = {
                'count': stats['count'],
                'min': int(stats['min']),
                'max': int(stats['max']),
                'avg': int(stats['avg']),
                'sum': int(stats['sum']),
                'first_hit': first,
                'last_hit': last
            }

        self.ds.signature.update(p_id, [
            (self.ds.signature.UPDATE_SET, 'stats', up_stats)
        ])

        return up_stats

    @elasticapm.capture_span(span_type='datastore')
    def calculate_signature_stats(self):
        sig_list = sorted([(x['id'], x['source'], x['name'], x['type'])
                           for x in self.ds.signature.stream_search("id:*",
                                                                    fl="id,name,type,source,classification",
                                                                    as_obj=False)])

        with concurrent.futures.ThreadPoolExecutor(max(min(len(sig_list), 20), 1)) as executor:
            for sid, source, name, sig_type in sig_list:
                executor.submit(self.get_stat_for_signature, sid, source, name, sig_type)

    @elasticapm.capture_span(span_type='datastore')
    def list_all_services(self, as_obj=True, full=False) -> Union[List[dict], List[Service]]:
        """
        :param as_obj: Return ODM objects rather than dicts
        :param full: If true retrieve all the fields of the service object, otherwise only
                     fields returned by search are given.
        """
        mask = None if full else list(self.ds.service.stored_fields.keys())

        # List all services from service delta (Return all fields if full is true)
        service_delta = list(self.ds.service_delta.stream_search("id:*", fl="*" if full else None))

        # Gather all matching services and apply a mask if we don't want the full source object
        service_data = [Service(s, mask=mask)
                        for s in self.ds.service.multiget([f"{item.id}_{item.version}" for item in service_delta],
                                                          as_obj=False, as_dictionary=False)]

        # Recursively update the service data with the service delta while stripping nulls
        services = [recursive_update(data.as_primitives(strip_null=True), delta.as_primitives(strip_null=True),
                                     stop_keys=['config'])
                    for data, delta in zip(service_data, service_delta)]

        # Return as an objet if needs be...
        if as_obj:
            return [Service(s, mask=mask) for s in services]
        else:
            return services

    @elasticapm.capture_span(span_type='datastore')
    def list_all_heuristics(self, as_obj=True) -> Union[List[dict], List[Heuristic]]:
        """
        :param as_obj: Return ODM objects rather than dicts
        """
        heuristics = list(self.ds.heuristic.stream_search("id:*", as_obj=as_obj))
        return heuristics

    @elasticapm.capture_span(span_type='datastore')
    def save_or_freshen_file(self, sha256, fileinfo, expiry, classification,
                             cl_engine=forge.get_classification(), redis=None, is_section_image=False):
        while True:
            current_fileinfo, version = self.ds.file.get_if_exists(
                sha256, as_obj=False, archive_access=config.datastore.ilm.update_archive, version=True)

            if current_fileinfo is None:
                current_fileinfo = {}

            # Remove control fields from file info and update current file info
            for x in ['classification', 'expiry_ts', 'seen', 'archive_ts']:
                fileinfo.pop(x, None)
            current_fileinfo.update(fileinfo)

            current_fileinfo['archive_ts'] = now_as_iso(config.datastore.ilm.days_until_archive * 24 * 60 * 60)

            # Update expiry time
            if isinstance(expiry, datetime):
                expiry = expiry.strftime(DATEFORMAT)
            current_expiry = current_fileinfo.get('expiry_ts', expiry)
            if current_expiry and expiry:
                current_fileinfo['expiry_ts'] = max(current_expiry, expiry)
            else:
                current_fileinfo['expiry_ts'] = None

            # Update seen counters
            now = now_as_iso()
            current_fileinfo['seen'] = seen = current_fileinfo.get('seen', {})
            seen['count'] = seen.get('count', 0) + 1
            seen['last'] = now
            seen['first'] = seen.get('first', now)

            # Update Classification
            classification = cl_engine.min_classification(
                str(current_fileinfo.get('classification', classification)),
                str(classification)
            )
            current_fileinfo['classification'] = classification

            # Update section image status
            current_fileinfo['is_section_image'] = current_fileinfo.get('is_section_image', False) or is_section_image

            try:
                self.ds.file.save(sha256, current_fileinfo, version=version)
                return
            except VersionConflictException as vce:
                log.info(f"Retrying save or freshen due to version conflict: {str(vce)}")



from typing import List, Dict, Optional as Opt
from assemblyline import odm
from assemblyline.odm.models.submission import SubmissionParams, File, Submission as DatabaseSubmission

MSG_TYPES = {"SubmissionIngested", "SubmissionReceived", "SubmissionStarted", "SubmissionCompleted"}
LOADER_CLASS = "assemblyline.odm.messages.submission.SubmissionMessage"


@odm.model(index=True, store=True)
class Notification(odm.Model):
    queue = odm.Optional(odm.Keyword())      # Queue to publish the completion message
    threshold = odm.Optional(odm.Integer())  # Notify only if this score threshold is met


@odm.model()
class Submission(odm.Model):
    sid = odm.UUID()                                                     # Submission ID to use
    time = odm.Date(default="NOW")
    files: List[File] = odm.List(odm.Compound(File), default=[])         # File block
    metadata: Dict[str, str] = odm.FlattenedObject(default={})           # Metadata submitted with the file
    notification: Notification = odm.Compound(Notification, default={})  # Notification queue parameters
    params: SubmissionParams = odm.Compound(SubmissionParams)            # Parameters of the submission
    scan_key: Opt[str] = odm.Optional(odm.Keyword())


def from_datastore_submission(submission: DatabaseSubmission):
    """
    A helper to convert between database model version of Submission
    and the message version of Submission.
    """
    return Submission({
        'sid': submission.sid,
        'files': submission.files,
        'metadata': submission.metadata,
        'params': submission.params,
        'scan_key': submission.scan_key
    })


@odm.model()
class SubmissionMessage(odm.Model):
    msg = odm.Compound(Submission)                                       # Body of the message
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS)   # Class to use to load the message as an object
    msg_type = odm.Enum(values=MSG_TYPES)                                # Type of message
    sender = odm.Keyword()                                               # Sender of the message




from assemblyline import odm
from assemblyline.common.caching import generate_conf_key

STATUSES = {"FAIL_NONRECOVERABLE", "FAIL_RECOVERABLE"}
ERROR_TYPES = {
    "UNKNOWN": 0,
    "EXCEPTION": 1,
    "MAX DEPTH REACHED": 10,
    "MAX FILES REACHED": 11,
    "MAX RETRY REACHED": 12,
    "SERVICE BUSY": 20,
    "SERVICE DOWN": 21,
    "TASK PRE-EMPTED": 30
}


@odm.model(index=True, store=True)
class Response(odm.Model):
    message = odm.Text(copyto="__text__")                                # Error message
    service_debug_info = odm.Optional(odm.Keyword())                     # Info about where the service was processed
    service_name = odm.Keyword(copyto="__text__")                        # Name of the service that had an error
    service_tool_version = odm.Optional(odm.Keyword(copyto="__text__"))  # Tool version of the service
    service_version = odm.Keyword()                                      # Version of the service
    status = odm.Enum(values=STATUSES)                                   # Status of the error


@odm.model(index=True, store=True)
class Error(odm.Model):
    archive_ts = odm.Date(store=False)                                     # Archiving timestamp
    created = odm.Date(default="NOW")                                      # Date at which the error was created
    expiry_ts = odm.Optional(odm.Date(store=False))                        # Expiry time stamp
    response: Response = odm.Compound(Response)                            # Response from the service
    sha256 = odm.SHA256(copyto="__text__")                                 # Hash of the file the error is related to
    type = odm.Enum(values=list(ERROR_TYPES.keys()), default="EXCEPTION")  # Type of error

    def build_key(self, service_tool_version=None, task=None):
        key_list = [
            self.sha256,
            self.response.service_name.replace('.', '_'),
            f"v{self.response.service_version.replace('.', '_')}",
            f"c{generate_conf_key(service_tool_version=service_tool_version, task=task)}",
            f"e{ERROR_TYPES.get(self.type, 0)}"]

        return '.'.join(key_list)



from assemblyline import odm


@odm.model(index=True, store=True)
class Statistics(odm.Model):
    count = odm.Integer(default=0)
    min = odm.Integer(default=0)
    max = odm.Integer(default=0)
    avg = odm.Integer(default=0)
    sum = odm.Integer(default=0)
    first_hit = odm.Optional(odm.Date())
    last_hit = odm.Optional(odm.Date())



import json
from typing import Generic, TypeVar, Optional
from collections import defaultdict

from assemblyline.remote.datatypes import get_client, retry_call

T = TypeVar('T')


class NamedQueue(Generic[T]):
    def __init__(self, name: str, host=None, port=None, private: bool = False, ttl: int = 0):
        self.c = get_client(host, port, private)
        self.name: str = name
        self.ttl: int = ttl

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.delete()

    def delete(self):
        retry_call(self.c.delete, self.name)

    def __len__(self):
        return self.length()

    def length(self):
        return retry_call(self.c.llen, self.name)

    def peek_next(self) -> Optional[T]:
        response = retry_call(self.c.lrange, self.name, 0, 0)

        if response:
            return json.loads(response[0])
        return None

    def pop(self, blocking: bool = True, timeout: int = 0) -> Optional[T]:
        if blocking:
            response = retry_call(self.c.blpop, self.name, timeout)
        else:
            response = retry_call(self.c.lpop, self.name)

        if not response:
            return response

        if blocking:
            return json.loads(response[1])
        else:
            return json.loads(response)

    def push(self, *messages: T):
        for message in messages:
            retry_call(self.c.rpush, self.name, json.dumps(message))
        if self.ttl:
            retry_call(self.c.expire, self.name, self.ttl)

    def unpop(self, *messages: T):
        """Put all messages passed back at the head of the FIFO queue."""
        for message in messages:
            retry_call(self.c.lpush, self.name, json.dumps(message))
        if self.ttl:
            retry_call(self.c.expire, self.name, self.ttl)


def select(*queues, **kw):
    """
    Do a blocking pop on each queue in sequence.

    Queues that map to the same keyslot will get bundled into the same call.

    This isn't as efficent as it was on a single node. Consider carefully
    the options if you need efficent select over many queues, this probably isn't it.
    """

    timeout = kw.get('timeout', 0)
    if len(queues) < 1:
        raise TypeError('At least one queue must be specified')
    if any([type(q) != NamedQueue for q in queues]):
        raise TypeError('Only NamedQueues supported')

    client = queues[0].c

    if hasattr(client.connection_pool, 'nodes'):
        mapping = defaultdict(list)
        for queue in queues:
            keyslot = client.connection_pool.nodes.keyslot(queue.name)
            mapping[keyslot].append(queue)
    else:
        mapping = {'': queues}

    if len(mapping) == 1:
        response = retry_call(client.blpop, [q.name for q in queues], timeout)

        if not response:
            return response

        return response[0].decode('utf-8'), json.loads(response[1])
    else:
        total = 0
        while total <= timeout or timeout == 0:
            for node_queues in mapping.values():
                total += 1
                response = retry_call(client.blpop, [q.name for q in node_queues], 1)

                if not response:
                    continue

                return response[0].decode('utf-8'), json.loads(response[1])
        return None



import binascii

from assemblyline.common.chunk import chunk

FILTER = b''.join([bytes([x]) if x in range(32, 127) else b'.' for x in range(256)])


def dump(binary: bytes, size: int = 2, sep: bytes = b" ") -> bytes:
    hexstr = binascii.hexlify(binary)
    return sep.join(chunk(hexstr, size))


def hexdump(binary: bytes, length: int = 16, indent: str = "", indent_size: int = 0, newline: str = '\n',
            prefix_offset: int = 0) -> str:
    """
    Create a string buffer that shows the given data in hexdump format.

        src -> source buffer
        length = 16 -> number of bytes per line
        indent = "" -> indentation before each lines
        indent_size = 0 -> number of time to repeat that indentation
        newline = "\n" -> chars used as newline char

        Example of output:
        00000000:  48 54 54 50 2F 31 2E 31 20 34 30 34 20 4E 6F 74  HTTP/1.1 404 Not
        00000010:  20 46 6F 75 6E 64 0D 0A 43 6F 6E 74              Found..Cont
        ...
    """
    generator = chunk(binary, length)
    line_frmt = "%%s%%08X:  %%-%ss   %%s" % ((length * 3) - 1)

    out = [line_frmt % (indent * indent_size, prefix_offset + (addr * length), dump(d).decode(),
                        d.translate(FILTER).decode())
           for addr, d in enumerate(generator)]
    return newline.join(out)




import re
from typing import AnyStr, Optional

from assemblyline.common import forge
from assemblyline.common.isotime import now_as_iso
from assemblyline.filestore import FileStore

DEFAULT_CACHE_LEN = 60 * 60  # 1 hour
COMPONENT_VALIDATOR = re.compile("^[a-zA-Z0-9][a-zA-Z0-9_.]*$")


class CacheStore(object):
    def __init__(self, component: str, config=None, datastore=None):
        if not component:
            raise ValueError("Cannot instantiate a cachestore without providing a component name.")

        if not COMPONENT_VALIDATOR.match(component):
            raise ValueError("Invalid component name. (Only letters, numbers, underscores and dots allowed)")

        if config is None:
            config = forge.get_config()

        self.component = component
        self.datastore = datastore or forge.get_datastore(config=config)
        self.filestore = FileStore(*config.filestore.cache)

    def __enter__(self) -> 'CacheStore':
        return self

    def __exit__(self, ex_type, exc_val, exc_tb):
        self.filestore.close()

    def save(self, cache_key: str, data: AnyStr, ttl=DEFAULT_CACHE_LEN, force=False):
        if not COMPONENT_VALIDATOR.match(cache_key):
            raise ValueError("Invalid cache_key for cache item. "
                             "(Only letters, numbers, underscores and dots allowed)")

        new_key = f"{self.component}_{cache_key}" if self.component else cache_key

        self.datastore.cached_file.save(new_key, {'expiry_ts': now_as_iso(ttl), 'component': self.component})
        self.filestore.put(new_key, data, force=force)

    def upload(self, cache_key: str, path: str, ttl=DEFAULT_CACHE_LEN):
        if not COMPONENT_VALIDATOR.match(cache_key):
            raise ValueError("Invalid cache_key for cache item. "
                             "(Only letters, numbers, underscores and dots allowed)")

        new_key = f"{self.component}_{cache_key}" if self.component else cache_key

        self.datastore.cached_file.save(new_key, {'expiry_ts': now_as_iso(ttl), 'component': self.component})
        self.filestore.upload(new_key, path, force=True)

    def touch(self, cache_key: str, ttl=DEFAULT_CACHE_LEN):
        if not COMPONENT_VALIDATOR.match(cache_key):
            raise ValueError("Invalid cache_key for cache item. "
                             "(Only letters, numbers, underscores and dots allowed)")
        if not self.exists(cache_key):
            raise KeyError(cache_key)

        new_key = f"{self.component}_{cache_key}" if self.component else cache_key
        self.datastore.cached_file.save(new_key, {'expiry_ts': now_as_iso(ttl), 'component': self.component})

    def get(self, cache_key: str) -> Optional[bytes]:
        new_key = f"{self.component}_{cache_key}" if self.component else cache_key
        return self.filestore.get(new_key)

    def download(self, cache_key: str, path: str):
        new_key = f"{self.component}_{cache_key}" if self.component else cache_key
        return self.filestore.download(new_key, path)

    def exists(self, cache_key: str):
        new_key = f"{self.component}_{cache_key}" if self.component else cache_key
        return self.filestore.exists(new_key)

    def delete(self, cache_key: str, db_delete=True):
        new_key = f"{self.component}_{cache_key}" if self.component else cache_key

        self.filestore.delete(new_key)
        if db_delete:
            self.datastore.cached_file.delete(new_key)



import enum
import os
from assemblyline.common.path import modulepath


SUBMISSION_QUEUE = 'dispatch-submission-queue'
DISPATCH_TASK_HASH = 'dispatch-active-submissions'
DISPATCH_RUNNING_TASK_HASH = 'dispatch-active-tasks'
SCALER_TIMEOUT_QUEUE = 'scaler-timeout-queue'


# Some pure functions for getting queue lengths (effectively for appending/prepending constants to strings)
def service_queue_name(service: str) -> str:
    """Take the name of a service, and provide the queue name to send tasks to that service."""
    return 'service-queue-' + service


def make_watcher_list_name(sid: str) -> str:
    """Get the name of the list dispatcher will pull for sending out submission events."""
    return 'dispatch-watcher-list-' + sid


def get_temporary_submission_data_name(sid: str, file_hash: str) -> str:
    """The HashMap used for tracking auxiliary processing data."""
    return '/'.join((sid, file_hash, 'temp_data'))


def get_tag_set_name(sid: str, file_hash: str) -> str:
    """The HashSet used to track the tags for an in-process file."""
    return '/'.join((sid, file_hash, 'tags'))


# A table storing information about the state of a service, expected type is ExpiringHash
# with a default ttl of None, and the ttl set per field based on the timeouts of queries
# and service operation
class ServiceStatus(enum.IntEnum):
    Idle = 0
    Running = 1


SERVICE_STATE_HASH = 'service-stasis-table'

# A null empty accepts, accepts all. A null rejects, rejects nothing
DEFAULT_SERVICE_ACCEPTS = ".*"
DEFAULT_SERVICE_REJECTS = "empty|metadata/.*"

# Queue priority values for each bucket in the ingester
PRIORITIES = {
    'low': 100,  # 0 -> 100
    'medium': 200,  # 101 -> 200
    'high': 300,
    'critical': 400,
    'user-low': 500,
    'user-medium': 1000,
    'user-high': 1500
}
MAX_PRIORITY = 2000

# The above priority values presented as a range for consistency
PRIORITY_RANGES = {}
_start = -1
for _end, _level in sorted((val, key) for key, val in PRIORITIES.items()):
    PRIORITY_RANGES[_level] = (_start + 1, _end)
    _start = _end


# Score thresholds for determining which queue priority a reingested item
# gets based on its previous score.
# eg.: item with a previous score of 99 will get 'low' priority
#      item with a previous score of 300 will get a 'high' priority
PRIORITY_THRESHOLDS = {
    'critical': 500,
    'high': 100,
}

RECOGNIZED_TYPES = {
    'android/apk': True,
    'android/dex': True,
    'android/resource': True,
    'android/xml': True,
    'archive/7-zip': True,
    'archive/ace': True,
    'archive/ar': True,
    'archive/audiovisual/flash': True,
    'archive/bzip2': True,
    'archive/cabinet': True,
    'archive/cart': True,
    'archive/chm': True,
    'archive/cpio': True,
    'archive/gzip': True,
    'archive/iso': True,
    'archive/lzma': True,
    'archive/rar': True,
    'archive/tar': True,
    'archive/tnef': True,
    'archive/unknown': True,
    'archive/xz': True,
    'archive/zip': True,
    'audiovisual/afs': True,
    'audiovisual/acb': True,
    'audiovisual/flash': True,
    'audiovisual/fsb': True,
    'audiovisual/unknown': True,
    'certificate/rsa': True,
    'code/animation/ccb': True,
    'code/autorun': True,
    'code/batch': True,
    'code/c': True,
    'code/css': True,
    'code/csharp': True,
    'code/erlang': True,
    'code/gles': True,
    'code/glsl': True,
    'code/go': True,
    'code/hta': True,
    'code/html': True,
    'code/java': True,
    'code/javascript': True,
    'code/jscript': True,
    'code/ida': True,
    'code/lisp': True,
    'code/nu': True,
    'code/pdfjs': True,
    'code/perl': True,
    'code/php': True,
    'code/ps1': True,
    'code/python': True,
    'code/ruby': True,
    'code/rust': True,
    'code/shell': True,
    'code/sgml': True,
    'code/sql': True,
    'code/vbe': True,
    'code/vbs': True,
    'code/wsf': True,
    'code/xml': True,
    'db/dbf': True,
    'db/sqlite': True,
    'document/installer/windows': True,
    'document/office/equation': True,
    'document/office/excel': True,
    'document/office/email': True,
    'document/office/mhtml': True,
    'document/office/paintbrush': True,
    'document/office/package': True,
    'document/office/powerpoint': True,
    'document/office/passwordprotected': True,
    'document/office/ole': True,
    'document/office/onenote': True,
    'document/office/rtf': True,
    'document/office/unknown': True,
    'document/office/visio': True,
    'document/office/word': True,
    'document/office/wordpro': True,
    'document/office/wordperfect': True,
    'document/email': True,
    'document/pdf': True,
    'document/unknown': True,
    'executable/unknown': True,
    'executable/mach-o': True,
    'executable/windows/com': True,
    'executable/windows/dos': True,
    'executable/windows/pe': True,
    'executable/windows/pe32': True,
    'executable/windows/pe64': True,
    'executable/windows/dll32': True,
    'executable/windows/dll64': True,
    'executable/linux/elf32': True,
    'executable/linux/elf64': True,
    'executable/linux/so32': True,
    'executable/linux/so64': True,
    'font/texture/pvr': True,
    'image/bmp': True,
    'image/gif': True,
    'image/jpg': True,
    'image/png': True,
    'image/svg': True,
    'image/texture/ka3d': True,
    'image/texture/ktx': True,
    'image/texture/pkm': True,
    'image/texture/powervr': True,
    'image/texture/rvio': True,
    'image/unknown': True,
    'ios/ipa': True,
    'java/class': True,
    'java/jar': True,
    'java/java': True,
    'java/jbdiff': True,
    'java/manifest': True,
    'java/signature': True,
    'java/unknown': True,
    'meta/torrent': True,
    'meta/shortcut/windows': True,
    'metadata/memorydump': True,
    'metadata/sysmon': True,
    'network/sff': True,
    'network/tcpdump': True,
    'network/unknown': True,
    'resource/big': True,
    'resource/ccz': True,
    'resource/cpk': True,
    'resource/dz': True,
    'resource/pak': True,
    'resource/ptc': True,
    'resource/sbm': True,
    'resource/sbr': True,
    'resource/sc': True,
    'resource/unity': True,
    'text/calendar': True,
    'text/json': True,
    'text/markdown': True,
    'quarantine/ahnlab': True,
    'quarantine/avast': True,
    'quarantine/avira': True,
    'quarantine/mcafee': True,
    'quarantine/windowsdefender': True,
    'unknown': True,
}

custom_rules = os.path.join(modulepath(__name__), 'custom.magic')
RULE_PATH = ':'.join((custom_rules, '/usr/share/file/magic.mgc'))



from assemblyline import odm

MSG_TYPES = {"MetricsCounter"}
LOADER_CLASS = "assemblyline.odm.messages.metrics.MetricsMessage"


@odm.model()
class Metrics(odm.Model):
    host = odm.Keyword()
    type = odm.Keyword()
    name = odm.Keyword()
    metrics = odm.Mapping(odm.Integer())


@odm.model()
class MetricsMessage(odm.Model):
    msg = odm.Compound(Metrics)
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS)
    msg_type = odm.Enum(values=MSG_TYPES, default="MetricsCounter")
    sender = odm.Keyword()



import logging
import os
import time
from typing import Optional

# noinspection PyProtectedMember
from azure.core.exceptions import ServiceRequestError, DecodeError, \
    ResourceExistsError, ResourceNotFoundError, ClientAuthenticationError, \
    ResourceModifiedError, ResourceNotModifiedError, TooManyRedirectsError, ODataV4Error
from azure.storage.blob import BlobServiceClient
from io import BytesIO

from assemblyline.common.exceptions import ChainAll
from assemblyline.filestore.transport.base import Transport, TransportException


"""
This class assumes a flat file structure in the Azure storage blob.
"""


@ChainAll(TransportException)
class TransportAzure(Transport):

    def __init__(self, base=None, access_key=None, host=None, connection_attempts=None):
        self.log = logging.getLogger('assemblyline.transport.azure')
        self.read_only = False
        self.connection_attempts: Optional[int] = connection_attempts

        # Data
        self.blob_container = base.strip("/")
        self.access_key = access_key
        self.host = host
        self.endpoint_url = f"https://{self.host}"

        # Clients
        self.service_client = BlobServiceClient(account_url=self.endpoint_url, credential=self.access_key)
        self.container_client = self.service_client.get_container_client(self.blob_container)

        # Init
        try:
            self.with_retries(self.container_client.get_container_properties)
        except TransportException as e:
            if not isinstance(e.cause, ResourceNotFoundError):
                raise
            try:
                self.with_retries(self.container_client.create_container)
            except TransportException as error:
                if not isinstance(error.cause, ResourceNotFoundError):
                    raise
                self.log.info('Failed to create container, we\'re most likely in read only mode')
                self.read_only = True

        def azure_normalize(path):
            # flatten path to just the basename
            return os.path.basename(path)

        super(TransportAzure, self).__init__(normalize=azure_normalize)

    def __str__(self):
        return f"azure://{self.host}/{self.blob_container}/"

    def with_retries(self, func, *args, **kwargs):
        retries = 0
        while self.connection_attempts is None or retries <= self.connection_attempts:
            try:
                ret_val = func(*args, **kwargs)

                if retries:
                    self.log.info('Reconnected to Azure transport!')

                return ret_val

            except (ServiceRequestError, DecodeError, ResourceExistsError, ResourceNotFoundError,
                    ClientAuthenticationError, ResourceModifiedError, ResourceNotModifiedError,
                    TooManyRedirectsError, ODataV4Error):
                # These errors will be wrapped by TransportException
                raise

            except Exception as e:
                self.log.warning(f"No connection to Azure transport "
                                 f"{os.path.join(self.endpoint_url, self.blob_container)}, retrying... "
                                 f"[{e.__class__.__name__}: {str(e)}]")
                time.sleep(0.25)
                retries += 1
        raise ConnectionError(f"Couldn't reach the requested azure endpoint {self.endpoint_url} inside retry limit")

    def delete(self, path):
        if self.read_only:
            raise TransportException("READ ONLY TRANSPORT: Method not allowed")

        key = self.normalize(path)
        blob_client = self.service_client.get_blob_client(self.blob_container, key)
        self.with_retries(blob_client.delete_blob)

    def exists(self, path):
        key = self.normalize(path)
        blob_client = self.service_client.get_blob_client(self.blob_container, key)
        try:
            blob_client.get_blob_properties()
        except ResourceNotFoundError:
            return False

        return True

    def makedirs(self, path):
        # Does not need to do anything as azurestorage blob has a flat layout.
        pass

    # File based functions
    def download(self, src_path, dst_path):
        key = self.normalize(src_path)
        dir_path = os.path.dirname(dst_path)

        # create dst_path if it doesn't exist
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        # download the key from azure
        with open(dst_path, "wb") as my_blob:
            blob_client = self.service_client.get_blob_client(self.blob_container, key)
            blob_data = self.with_retries(blob_client.download_blob)
            blob_data.readinto(my_blob)

    def upload(self, src_path, dst_path):
        if self.read_only:
            raise TransportException("READ ONLY TRANSPORT: Method not allowed")

        key = self.normalize(dst_path)

        # if file exists already, it will be overwritten
        with open(src_path, "rb") as data:
            blob_client = self.service_client.get_blob_client(self.blob_container, key)
            try:
                self.with_retries(blob_client.upload_blob, data, overwrite=True)
            except TransportException as error:
                if not isinstance(error.cause, ResourceExistsError):
                    raise

    # Buffer based functions
    def get(self, path: str) -> bytes:
        key = self.normalize(path)
        my_blob = BytesIO()

        blob_client = self.service_client.get_blob_client(self.blob_container, key)
        blob_data = self.with_retries(blob_client.download_blob)
        blob_data.readinto(my_blob)
        return my_blob.getvalue()

    def put(self, dst_path, content):
        if self.read_only:
            raise TransportException("READ ONLY TRANSPORT: Method not allowed")

        key = self.normalize(dst_path)
        if isinstance(content, str):
            content = content.encode('utf-8')

        with BytesIO(content) as file_io:
            blob_client = self.service_client.get_blob_client(self.blob_container, key)
            try:
                self.with_retries(blob_client.upload_blob, file_io, overwrite=True)
            except TransportException as error:
                if not isinstance(error.cause, ResourceExistsError):
                    raise



from assemblyline import odm
from assemblyline.common import forge

Classification = forge.get_classification()
SAFEHASH_TYPES = ["file", "tag", "signature"]
SOURCE_TYPES = ["user", "external"]


@odm.model(index=True, store=True)
class Hashes(odm.Model):
    md5 = odm.Optional(odm.MD5(copyto="__text__"))        # MD5 hash
    sha1 = odm.Optional(odm.SHA1(copyto="__text__"))      # SHA1 Hash
    sha256 = odm.Optional(odm.SHA256(copyto="__text__"))  # SHA256 Hash


@odm.model(index=True, store=False)
class File(odm.Model):
    name = odm.List(odm.Keyword(store=True, copyto="__text__"), default=[])  # List of names seen for that file
    size = odm.Optional(odm.Integer())                                       # Size of the file
    type = odm.Optional(odm.Keyword())                                       # Type of file


@odm.model(index=True, store=False)
class Source(odm.Model):
    classification = odm.Classification(default=Classification.UNRESTRICTED)  # Classification of the source
    name = odm.Keyword(store=True)                                            # Name of the source
    reason = odm.List(odm.Keyword())                                          # Reason why it is safelisted
    type = odm.Enum(values=SOURCE_TYPES)                                      # Type of source


@odm.model(index=True, store=True)
class Tag(odm.Model):
    type = odm.Keyword()                                                      # List of names seen for that file
    value = odm.Keyword(copyto="__text__")                                    # Size of the file


@odm.model(index=True, store=True)
class Signature(odm.Model):
    name = odm.Keyword(copyto="__text__")                                     # Name of the signature


@odm.model(index=True, store=True)
class Safelist(odm.Model):
    added = odm.Date(default="NOW")                       # Date when the safe hash was added
    classification = odm.Classification()                 # Computed max classification for the safe hash
    enabled = odm.Boolean(default=True)                   # The safe hash enabled or not
    hashes = odm.Compound(Hashes, default={})             # List of hashes related to the safe hash
    file = odm.Optional(odm.Compound(File))               # Informations about the file
    sources = odm.List(odm.Compound(Source))              # List of reasons why hash is safelisted
    tag = odm.Optional(odm.Compound(Tag))                 # Informations about the tag
    signature = odm.Optional(odm.Compound(Signature))     # Informations about the signature
    type = odm.Enum(values=SAFEHASH_TYPES)                # Type of safe hash
    updated = odm.Date(default="NOW")                     # Last date when sources were added to the safe hash


if __name__ == "__main__":
    from pprint import pprint
    from assemblyline.odm.randomizer import random_model_obj
    pprint(random_model_obj(Safelist, as_json=True))



from assemblyline import odm

OS_TYPES = ["windows", "linux"]
OS_VARIANTS_WINDOWS = ["winxp", "win7", "win10"]
OS_VARIANTS_LINUX = ["ubuntuprecise", "ubuntutrusty", "ubuntuxenial", "ubuntubionic"]
OS_VARIANTS = OS_VARIANTS_WINDOWS + OS_VARIANTS_LINUX


@odm.model(index=True, store=True)
class VM(odm.Model):
    enabled = odm.Boolean(default=True)
    name = odm.Keyword(copyto="__text__")
    num_workers = odm.Integer(default=1)
    os_type = odm.Enum(values=OS_TYPES)
    os_variant = odm.Enum(values=OS_VARIANTS)
    ram = odm.Integer(default=1024)
    revert_every = odm.Integer(default=600)
    vcpus = odm.Integer(default=1)
    virtual_disk_url = odm.Keyword()
